# %%
from __future__ import annotations

import warnings
from math import ceil
from pathlib import Path

import fastsim
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import gaussian_kde
import os
from t3co.run import Global as gl
from t3co.run import run_scenario

# TODO: think about this stuff
# Thoughts about how to improve this:
# 1 - maybe $/lb should be in terms of profit and not raw shipping fees
# 2 - could consider scaling number of vehicles by some function of opportunity cost
#     such that fungible goods are shipped at same rate.  This would bump up capital
#     expenses but lower lost revenue.  At the other end, we could just lose some revenue.
#     I would think they'd work out pretty similarly, but it'd be interesting to compare
#     two approaches.


class OpportunityCost:
    """
    This class is used to calculate the different opportunity costs for a scenario and vehicle
    - Payload Capacity Cost Multiplier
    - Fueling Downtime Cost
    - Maintenance and Repair Downtime Cost
    """

    def __init__(
        self, scenario: run_scenario.Scenario, range_dict: dict = None, **kwargs
    ) -> None:
        """
        Initializes OpportunityCost object using Scenario object, range_dict (from fueleconomy module), and other arguments

        Args:
            scenario (run_scenario.Scenario): Scenario object
            range_dict (dict, optional): dictionary containing primary_fuel_range_mi from fueleconomy.get_range_mi function. Defaults to None.
        """

        # maybe this should be on a profit basis (it's currently not)
        ### All params below here are experimental and not to be used for now
        # self.cost_per_lb_mi = kwargs.pop("cost_per_lb_mi", 0.0003)
        # trip distance[mi] for typical point A to point B trip
        # self.d_trip_mi = kwargs.pop('d_trip_mi', 500)
        # self.d_trip_mi = scenario.vmt[0]/scenario.shifts_per_year
        self.payload_cap_cost_multiplier = None
        # distance[mi] that can be traveled with a full battery.
        # self.total_range_mi = kwargs.pop('d_fullbatt_mi', 300)
        if range_dict:
            self.total_range_mi = range_dict["primary_fuel_range_mi"]
            # print('primary_fuel_range_mi')
        else:
            self.total_range_mi = scenario.target_range_mi
        # print(range_dict)
        # self.cycle_distance_mi = range_dict['cycle_distance_mi']
        # average vehicle speed[mph] over trip, should come from FASTSim
        # self.v_mean_mph = range_dict['mean_cyc_speed_mph']
        # driver hourly labor rate[$/hr].
        # self.free_dwell_time_hr = scenario.fdt_available_freetime_hr
        # dwell time [hr] driven by non-charging needs (e.g. required work breaks) that can be used for charging.
        # self.free_dwell_time_hr = kwargs.pop('free_dwell_time_hr', 3)
        # fraction of dwell time that can actually be used for charging
        # self.dwell_time_efficiency = kwargs.pop('dwell_time_efficiency', 1.)
        if scenario.activate_tco_fueling_dwell_time_cost:
            self.frac_of_fullcharge_bounds = list(
                np.float_(scenario.fdt_frac_full_charge_bounds.strip(" ][").split(","))
            )
            self.shifts_per_year = list(
                np.float_(scenario.shifts_per_year.strip(" ][").split(","))
            )

        self.payload_cap_cost_multiplier = 0
        
        # rate[kW] at which battery is recharged
        # self.charge_rate_kW = kwargs.pop("charge_rate_kW", 200)
        # battery capacity [kilowatt-hours]
        # self.battery_capacity_kWhr = kwargs.pop("batt_cap_kW_hr", 300)
        # number of full battery trips that don't incur dwell time penalty
        # battery is allowed `num_free_trips` charging events without penalty
        # scenario.fdt_num_free_dwell_trips = kwargs.pop('num_free_trips', 1.)
        # scenario.fdt_num_free_dwell_trips = scenario.fdt_num_free_dwell_trips
        # number of weight weight bins
        # self.num_bins = kwargs.pop("num_bins", 1_000)

        if len(kwargs) > 0:
            warnings.warn(f"Invalid kwargs: {list(kwargs.keys())}")

        # weight distribution file to load
        self.wt_dist_file = kwargs.pop(
            "wt_dist_file",
            Path(os.path.abspath(__file__)).parents[1]
            / "resources"
            / "auxiliary"
            / "tractorweightvars.csv",
        )
        self.df_veh_wt = pd.read_csv(self.wt_dist_file, index_col=0)

    def set_kdes(
        self,
        scenario: run_scenario.Scenario,
        bw_method: float = 0.15,
        verbose: bool = False,
    ) -> None:
        """
        This method sets tje kde kernel. This is time-consuming, only call this once, if possible.

        Args:
            scenario (run_scenario.Scenario): Scenario object
            bw_method (float, optional):  kernel bandwidth method used by guassian_kde. Defaults to .15.
            verbose (bool, optional): if True, prints process sets. Defaults to False.
        """

        if verbose:
            print("Initializing kernels.")

        self.df_veh_wt = self.df_veh_wt[~self.df_veh_wt["WEIGHTAVG"].isnull()]
        self.df_veh_wt = self.df_veh_wt[~self.df_veh_wt["WEIGHTEMPTY"].isnull()]
        self.df_veh_wt = self.df_veh_wt[self.df_veh_wt["WEIGHTAVG"] < 120000]

        weights = self.df_veh_wt["TAB_MILES"] / np.nansum(self.df_veh_wt["TAB_MILES"])
        kernel = gaussian_kde(
            self.df_veh_wt["WEIGHTAVG"], weights=weights, bw_method=bw_method
        )
        self.vehicle_weights_bins_lb = np.linspace(
            self.df_veh_wt["WEIGHTAVG"].min(), self.df_veh_wt["WEIGHTAVG"].max(), 1000
        )
        self.vehicle_weights_bins_kg = gl.lbs_to_kgs(self.vehicle_weights_bins_lb)

        # get probability of each vehicle weight
        self.p_of_weights = kernel(self.vehicle_weights_bins_lb)
        # self.p_of_weights_normalized =

        probability_payload = pd.DataFrame(
            [self.vehicle_weights_bins_kg, self.p_of_weights],
            index=["vehicle_weights_bins_kg", "p_of_weights"],
        ).T
        probability_payload.to_csv(
            Path(self.wt_dist_file).parents[0] / "payload_pdf.csv"
        )
        normalization_factor = probability_payload[
            probability_payload["vehicle_weights_bins_kg"].between(
                scenario.plf_ref_veh_empty_mass_kg, scenario.gvwr_kg
            )
        ]["p_of_weights"].sum()
        self.p_of_weights_normalized = self.p_of_weights / normalization_factor
        # print(f'plf_ref_veh_empty_mass_kg {scenario.plf_ref_veh_empty_mass_kg}')

    def set_payload_loss_factor(
        self,
        a_vehicle: fastsim.vehicle.Vehicle,
        scenario: run_scenario.Scenario,
        plots: bool = False,
        plots_dir: str = None,
    ) -> None:
        """
        This method runs teh kernel density estimation function set_kdes and calculates the payload capacity loss factor (payload_cap_cost_multiplier) \
            of the new vehicle compared to a conventional vehicle's reference empty weight.

        Args:
            a_vehicle (fastsim.vehicle): FASTSim vehicle object of the analysis vehicle
            scenario (run_scenario.Scenario): Scenario object of current selection
            plots (bool, optional): if True, creates histogram of KDE weight bins. Defaults to False.
            plots_dir (str, optional): output diretory for saving plot figure. Defaults to None.
        """
        self.set_kdes(scenario, verbose=False)

        new_empty_weight_kg = a_vehicle.veh_kg - a_vehicle.cargo_kg
        empty_increase_kg = new_empty_weight_kg - scenario.plf_ref_veh_empty_mass_kg
        new_cargo_cieling_kg = (
            scenario.gvwr_kg - empty_increase_kg + scenario.gvwr_credit_kg
        )
        # print(new_cargo_cieling_lb)

        # determine indices where lost cargo capacity is bounded in vehicle_weights
        # and get the corresponding indices for p_of_weights
        # based on current vehicle's new_cargo_cieling_lb and base_vehicle_gvwr_lb

        if empty_increase_kg >= scenario.gvwr_credit_kg:
            a = self.vehicle_weights_bins_kg - new_cargo_cieling_kg
            # print(a)
            minidx = (
                np.where(
                    self.vehicle_weights_bins_kg == a[a > 0][0] + new_cargo_cieling_kg
                )
            )[0][0]
            a = self.vehicle_weights_bins_kg - scenario.gvwr_kg
            maxidx = (
                np.where(self.vehicle_weights_bins_kg == a[a > 0][0] + scenario.gvwr_kg)
            )[0][0]
            # print(f'Vehicle bin limits: {self.vehicle_weights_bins_kg[minidx]}, {self.vehicle_weights_bins_kg[maxidx]}')

            # print( f' PDF bounds: {minidx}, {maxidx}' )
            # print( f' PDF weight bounds: {self.vehicle_weights_bins_lb[minidx]}, {self.vehicle_weights_bins_lb[maxidx]}' )

            estimated_lost_payload_per_bin_kg = self.p_of_weights_normalized[
                minidx:maxidx
            ] * (self.vehicle_weights_bins_kg[minidx:maxidx] - new_cargo_cieling_kg)
            # estimated_lost_payload_lb = sum(estimated_lost_payload_per_bin_lb)
            estimated_lost_payload_kg = np.trapz(estimated_lost_payload_per_bin_kg)
            # print(f'estimated_lost_payload_kg: {estimated_lost_payload_kg}')

            # payload cost multiplier
            self.payload_cap_cost_multiplier = 1 + estimated_lost_payload_kg / (
                scenario.gvwr_kg - new_empty_weight_kg + scenario.gvwr_credit_kg
            )

            scenario.estimated_lost_payload_kg = estimated_lost_payload_kg
        else:
            self.payload_cap_cost_multiplier = 1
        # recording final report data on vehicle empty weights and cargo capcities
        scenario.plf_scenario_vehicle_empty_kg = new_empty_weight_kg
        scenario.plf_reference_vehicle_cargo_capacity_kg = (
            scenario.gvwr_kg - scenario.plf_ref_veh_empty_mass_kg
        )
        scenario.plf_scenario_vehicle_cargo_capacity_kg = (
            scenario.gvwr_kg + scenario.gvwr_credit_kg - new_empty_weight_kg
        )

        def make_plots(save_dir: str = None) -> None:
            """
            This function generates a histogram of the payload KDE weight distribution

            Args:
                save_dir (str, optional): Output directory path to save plot figure. Defaults to None.
            """
            if save_dir and not Path(save_dir).exists():
                save_dir.mkdir()

            fig, ax = plt.subplots(figsize=(9, 6))
            ax.hist(
                gl.lbs_to_kgs(self.df_veh_wt["WEIGHTAVG"]),
                bins=50,
                label="WEIGHTAVG: operating weight [kg]",
            )
            ax2 = ax.twinx()
            ax2.plot(
                self.vehicle_weights_bins_kg,
                self.p_of_weights,
                color="red",
                linewidth=3,
                label="KDE",
            )
            ax2.fill_between(
                x=self.vehicle_weights_bins_kg[minidx:maxidx],
                y1=self.p_of_weights[minidx:maxidx],
                y2=[0] * len(self.p_of_weights[minidx:maxidx]),
                color="red",
                alpha=0.3,
                label=f"estimated cargo lost [kg]: {round(estimated_lost_payload_kg)}",
            )
            ax2.axvline(
                scenario.gvwr_kg,
                label=f"GVWR [kg] {round(scenario.gvwr_kg)}",
                color="orange",
            )
            ax2.axvline(
                new_cargo_cieling_kg,
                label=f"GVWR + credit - empty weight increase [kg] : {round(scenario.gvwr_kg)} + {round(scenario.gvwr_credit_kg)}  - {round(empty_increase_kg)} = {round(new_cargo_cieling_kg)}",
                color="purple",
            )
            fig.suptitle(
                f"payload cost multiplier: {round(self.payload_cap_cost_multiplier,2)}"
            )
            fig.legend()
            plt.show()

        if plots:
            make_plots(plots_dir)

    # Old Dwell time calculation
    # def get_dwell_loss_factor(self,
    #         a_vehicle:fastsim.vehicle,
    #         scenario) :
    #     self.dwell_time_hr = max(
    #     0, self.d_trip_mi / self.total_range_mi - scenario.fdt_num_free_dwell_trips) * \
    #     (a_vehicle.ess_max_kwh / a_vehicle.ess_max_kw)
    #     # (nominal time + additional non-driving time to charge) / nominal time

    #     # TODO: need to add separate labor and payload dwell penalties starting here
    #     self.net_fueling_dwell_time_hr_per_yr = max(
    #         0, (self.dwell_time_hr - max(0, self.free_dwell_time_hr)) * self.dwell_time_efficiency)

    #     # number of original trip time durations after correcting for EV time penalty
    #     # (e.g. 1.8 means 80% longer trip duration)
    #     self.dwell_time_factor = max(1, (self.d_trip_mi / self.v_mean_mph + self.net_fueling_dwell_time_hr_per_yr) /
    #         (self.d_trip_mi / self.v_mean_mph))

    def set_fueling_dwell_time_cost(
        self, a_vehicle: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario
    ) -> None:
        """
        This function calculates the fueling dwell time cost for a vehicle based on fuel fill rate/charging power and shifts_per_year

        Args:
            a_vehicle (fastsim.vehicle): FASTSim vehicle object of analysis vehicle
            scenario (run_scenario.Scenario): Scenario object for current selection
        """
        self.total_fueling_dwell_time_hr = 0
        self.net_fueling_dwell_time_hr_per_yr = []
        self.fueling_downtime_oppy_cost_dol_per_yr = []
        self.fueling_dwell_labor_cost_dol_per_yr = []
        dwellparams = np.array(
            [
                scenario.fdt_dwpt_fraction_power_pct,
                scenario.fdt_frac_full_charge_bounds,
                scenario.fdt_avg_overhead_hr_per_dwell_hr,
                scenario.downtime_oppy_cost_dol_per_hr,
            ]
        )
        assert any(
            dwellparams
        ), f"Missing parameters in {str(dwellparams)}: {np.isnan(dwellparams)}"

        if a_vehicle.veh_pt_type in ["BEV"]:
            self.full_dwell_hr = (1 - scenario.fdt_dwpt_fraction_power_pct) * (
                a_vehicle.ess_max_kwh / scenario.ess_max_charging_power_kw
            )
        elif a_vehicle.veh_pt_type in ["Conv"]:
            if scenario.fuel_type in ["gasoline"]:
                self.full_dwell_hr = (
                    a_vehicle.fs_kwh
                    / (gl.kwh_per_gge)
                    / scenario.fs_fueling_rate_gasoline_gpm
                ) / 60
            else:
                self.full_dwell_hr = (
                    a_vehicle.fs_kwh
                    / (gl.kwh_per_gge / gl.DieselGalPerGasGal)
                    / scenario.fs_fueling_rate_diesel_gpm
                ) / 60
        else:
            self.full_dwell_hr = (
                (1 - scenario.fdt_dwpt_fraction_power_pct)
                * (
                    a_vehicle.fs_kwh
                    / (gl.kwh_per_gge / gl.kgH2_per_gge)
                    / scenario.fs_fueling_rate_kg_per_min
                )
                / 60
            )

        for i in range(scenario.vehicle_life_yr):
            # self.shifts_per_year = (scenario.vmt[i])/ self.cycle_distance_mi
            self.d_trip_mi = scenario.vmt[i] / self.shifts_per_year[i]
            self.num_of_dwells = max(
                0,
                (
                    (self.d_trip_mi)
                    * (1 - scenario.fdt_dwpt_fraction_power_pct)
                    / self.total_range_mi
                    - scenario.fdt_num_free_dwell_trips
                ),
            )
            # print(f'num_of_dwells: {self.num_of_dwells}\n vmt: {scenario.vmt[i]}')
            # print(f'd_trip_mi: {self.d_trip_mi}')
            if self.num_of_dwells != 0:
                self.remaining_dwells = self.num_of_dwells % 1
                if self.remaining_dwells < self.frac_of_fullcharge_bounds[0]:
                    self.num_of_dwells += (
                        self.frac_of_fullcharge_bounds[0] - self.remaining_dwells
                    )
                elif (
                    self.frac_of_fullcharge_bounds[0]
                    < self.remaining_dwells
                    < self.frac_of_fullcharge_bounds[1]
                ):
                    self.num_of_dwells += 0
                else:
                    self.num_of_dwells += 1 - self.remaining_dwells
            # else:

            if (self.num_of_dwells < 1 and not scenario.fdt_num_free_dwell_trips) or (
                scenario.fuel_type
            ):
                self.net_fueling_dwell_time_hr_per_yr.append(
                    scenario.vmt[i]
                    * (1 - scenario.fdt_dwpt_fraction_power_pct)
                    / self.total_range_mi
                    * (self.full_dwell_hr + scenario.fdt_avg_overhead_hr_per_dwell_hr)
                )
            else:
                self.dwell_time_hr = (
                    self.num_of_dwells * self.full_dwell_hr
                    + ceil(self.num_of_dwells) * scenario.fdt_avg_overhead_hr_per_dwell_hr
                )
                self.net_fueling_dwell_time_hr_per_yr.append(
                    self.shifts_per_year[i]
                    * max(
                        0,
                        (
                            self.dwell_time_hr
                            - max(0, scenario.fdt_available_freetime_hr)
                        ),
                    )
                )

            self.fueling_dwell_labor_cost_dol_per_yr.append(self.net_fueling_dwell_time_hr_per_yr[i] * scenario.labor_rate_dol_per_hr)
            self.fueling_downtime_oppy_cost_dol_per_yr.append(self.net_fueling_dwell_time_hr_per_yr[i] * scenario.downtime_oppy_cost_dol_per_hr)
            
            self.total_fueling_dwell_time_hr += self.net_fueling_dwell_time_hr_per_yr[i]

        # self.dwell_time_factor = max(1, (self.d_trip_mi / self.v_mean_mph + self.net_fueling_dwell_time_hr_per_yr[i]) /
        #     (self.d_trip_mi / self.v_mean_mph))

    # def set_fueling_dwell_time_cost #Old calculation
    #           (self,
    #         a_vehicle:fastsim.vehicle,
    #         scenario):

    #     self.dwell_time_hr = 0 if self.d_trip_mi< self.total_range_mi else \
    #                           max((self.d_trip_mi - self.total_range_mi)* (a_vehicle.ess_max_kwh / a_vehicle.ess_max_kw), scenario.dlf_min_charge_time_hr)
    #     self.time_full_charge_hr = self.total_range_mi * a_vehicle.ess_max_kwh / a_vehicle.ess_max_kw
    #     self.time_frac_charge_hr = self.time_full_charge_hr *(1- (self.dwell_time_hr/self.time_full_charge_hr - floor(self.dwell_time_hr/self.time_full_charge_hr) ))
    #     self.dwell_time_hr = self.dwell_time_hr if self.time_frac_charge_hr < self.time_available_charge_hr else (self.dwell_time_hr+ self.time_frac_charge_hr)

    #     self.dwell_time_cost_Dol = self.dwell_time_hr/self.dwell_time_efficiency * self.labor_rate_dol_per_hr
    #     print(f'Dwell time hr per year: {self.dwell_time_hr}')
    #     print(f'Dwell cost dol: {self.dwell_time_cost_Dol}')

    def set_M_R_downtime_cost(
        self, a_vehicle: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario
    ) -> None:
        """
        This function calculates the Maintenance and Repair (M&R) downtime cost based on planned, unplanned, and tire replacement downtime inputs

        Args:
            a_vehicle (fastsim.vehicle): FASTSim object of the analysis vehicle
            scenario (run_scenario.Scenario): Scenario object for the current selection
        """
        self.planned_downtime_hr = [
            scenario.mr_planned_downtime_hr_per_yr
            for _ in range(scenario.vehicle_life_yr)
        ]  # regular maintenance and inspections
        self.unplanned_downtime_hr = [
            scenario.mr_unplanned_downtime_hr_per_mi[i] * scenario.vmt[i]
            for i in range(scenario.vehicle_life_yr)
        ]  # increases with age
        self.tire_replacement_downtime_hr = [
            (scenario.vmt[i])
            / scenario.mr_avg_tire_life_mi
            * scenario.mr_tire_replace_downtime_hr_per_event
            for i in range(scenario.vehicle_life_yr)
        ]
        self.net_net_mr_downtime_hr_per_yr_per_yr = np.array(
            [
                self.planned_downtime_hr[i]
                + self.unplanned_downtime_hr[i]
                + self.tire_replacement_downtime_hr[i]
                for i in range(scenario.vehicle_life_yr)
            ]
        )
        self.mr_downtime_oppy_cost_dol_per_yr = (
            self.net_net_mr_downtime_hr_per_yr_per_yr * scenario.downtime_oppy_cost_dol_per_hr
        )
        
        # print(f'Total M&R downtime hr: {self.net_net_mr_downtime_hr_per_yr}')
        # print(f'M&R Downtime cost dol: {self.net_MR_downtime_oppcosts_Dol}')


# %%
def main():
    """
    Runs the opportunity cost module as a standalone code based on input vehicles and scenarios
    """
    print("opportunity cost main()")
    vehicles_file = Path(
        "./resources/inputs/tda_example/TDA_FY22_vehicle_model_assumptions.csv"
    )
    scenarios_file = Path(
        "./resources/inputs/tda_example/TDA_FY22_scenario_assumptions.csv"
    )
    s, c = run_scenario.get_scenario_and_cycle(33, scenarios_file)
    v = run_scenario.get_vehicle(33, vehicles_file)
    oc = OpportunityCost(v, s)

    print("GVWRlb", gl.kg_to_lbs(s.gvwr_kg))
    print("GvwrCreditlb", gl.kg_to_lbs(s.gvwr_credit_kg))
    print("veh_lb", gl.kg_to_lbs(v.veh_kg))
    print("ess_mass_lb", gl.kg_to_lbs(v.ess_mass_kg))
    print("original empty lb", gl.kg_to_lbs(v.veh_kg - v.cargo_kg))
    run_scenario.set_max_battery_kwh(v, v.ess_max_kwh * 2)
    print("next ess_mass_lb", gl.kg_to_lbs(v.ess_mass_kg))
    print("next veh_lb", gl.kg_to_lbs(v.veh_kg))

    print("GVWRlb", oc.GVWRlb)
    print("gvwr_credit_lb", oc.gvwr_credit_lb)
    print("oc.base_vehicle_veh_lb", gl.kg_to_lbs(oc.base_vehicle_veh_kg))
    print("oc.base_vehicle_cargo_lb", gl.kg_to_lbs(oc.base_vehicle_cargo_kg))
    print("oc.original_empty_lb", gl.kg_to_lbs(oc.reference_vehicle_empty_kg))

    plf = oc.set_payload_loss_factor(v, plots=True)
    print(plf)
    oc.set_fueling_dwell_time_cost(v, s)
    print(oc.dwell_time_cost_Dol)
    print(oc.net_fueling_dwell_time_hr_per_yr)
    print(oc.__dict__["payload_cap_cost_multiplier"])
    oc.set_M_R_downtime_cost(v, s)


# %%
if __name__ == "__main__":
    oc = main()



# To add a new cell, type '# %%'
# To add a new markdown cell, type '# %% [markdown]'
# %%
from IPython import get_ipython

# %%
get_ipython().run_line_magic('matplotlib', 'inline')
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
import seaborn as sns

sns.set()

import importlib
from t3co.tco import opportunity_cost as oppcost
importlib.reload(oppcost)


# %%
free_dwells = np.linspace(0, 1.5, 15)
wt_xEV_diffs = np.linspace(-5, 5, 100) * 1_000

oc = oppcost.OpportunityCost()
oc.free_dwell_time_hr = 1

oppcosts = np.array([[{}] * len(wt_xEV_diffs)] * len(free_dwells))
for j, wt_xEV_diff in enumerate(wt_xEV_diffs):
    print(f'\nOuter iteration {j:d}')
    print(f'wt_xEV_Diff = {wt_xEV_diff:.0f} lbs')
    for i, free_dwell in enumerate(free_dwells):
        print(f'Inner iteration {i:d}')
        print(f'free_dwell = {free_dwell:.2f} hrs')
        oc.free_dwell_time_hr = free_dwell
        oc.wt_xEVempty_lb = oc.wt_base_vehicle_lb + wt_xEV_diff
        oppcosts[i, j] = oc.set_opp_cost()


# %%
tot_opp_cost_xy = np.array([[oppcosts[i][j]['tot_opp_cost_per_mi']
                    for i in range(len(free_dwells))] 
                   for j in range(len(wt_xEV_diffs))])

labor_cost_xy = np.array([[oppcosts[i][j]['labor_cost_per_mi']
                    for i in range(len(free_dwells))] 
                   for j in range(len(wt_xEV_diffs))])

payload_cost_xy = np.array([[oppcosts[i][j]['payload_cost_per_mi']
                    for i in range(len(free_dwells))] 
                   for j in range(len(wt_xEV_diffs))])

conv_payload_rev_xy = np.array([[oppcosts[i][j]['CV_rev__mi']
                    for i in range(len(free_dwells))] 
                   for j in range(len(wt_xEV_diffs))])

# conv_payload_rev_xy = np.array([[oppcosts[i][j]['xEV_rev__mi']
#                     for i in range(len(free_dwells))] 
#                    for j in range(len(wt_xEV_diffs))])

time_cost_xy = np.array([[oppcosts[i][j]['time_cost_per_mi']
                    for i in range(len(free_dwells))] 
                   for j in range(len(wt_xEV_diffs))])

payload_time_factor_xy = np.array([[oppcosts[i][j]['payload_time_factor']
                    for i in range(len(free_dwells))] 
                   for j in range(len(wt_xEV_diffs))])

net_dwell_time_xy = np.array([[oppcosts[i][j]['net_fueling_dwell_time_hr_per_yr']
                    for i in range(len(free_dwells))] 
                   for j in range(len(wt_xEV_diffs))])

dwell_time_xy = np.array([[oppcosts[i][j]['dwell_time_hr']
                    for i in range(len(free_dwells))] 
                   for j in range(len(wt_xEV_diffs))])


# %%
plt.plot(oc.weight_x_lb, oc.CV_cargo_kde, label='conventional')
plt.xlabel('Cargo Weight [lb]')
plt.ylabel('Probability Density')
plt.title('Cargo Weight Probability Density')
plt.legend()

# %% [markdown]
# ## Parametric plots

# %%
for i in range(len(free_dwells)):
    if (i % 4 == 0) or i == len(free_dwells) - 1:
        plt.plot(wt_xEV_diffs, tot_opp_cost_xy[:,i], 
                 label='{:.2f}'.format(free_dwells[i]))
    else:
        plt.plot(wt_xEV_diffs, tot_opp_cost_xy[:,i], linestyle='--')
plt.xlabel('xEV Wt. Minus CV Wt. [lb]')
plt.ylabel('Total Opportunity Cost [$/mi]')
plt.title('Total Opportunity Cost')
plt.legend(title='Free Dwell Time [hr]')


# %%
for i in range(len(free_dwells)):
    if (i % 4 == 0) or i == len(free_dwells) - 1:
        plt.plot(wt_xEV_diffs, (tot_opp_cost_xy / conv_payload_rev_xy)[:,i], 
                 label='{:.2f}'.format(free_dwells[i]))
    else:
        plt.plot(wt_xEV_diffs, (tot_opp_cost_xy / conv_payload_rev_xy)[:,i], linestyle='--')
plt.xlabel('xEV Wt. Minus CV Wt. [lb]')
plt.ylabel('Cost Ratio')
plt.title('Total Opportunity Cost\nPer Conventional Revenue')
plt.legend(title='Free Dwell Time [hr]')


# %%
# for i in range(len(free_dwells)):
#     if (i % 4 == 0) or i == len(free_dwells) - 1:
#         plt.plot(wt_xEV_diffs, payload_cost_xy[:,i], 
#                  label='{:.2f}'.format(free_dwells[i]))
#     else:
#         plt.plot(wt_xEV_diffs, payload_cost_xy[:,i], linestyle='--')
# plt.legend(title='Free Dwell Time [hr]')
plt.plot(wt_xEV_diffs, payload_cost_xy[:, 0])
plt.xlabel('xEV Wt. Minus CV Wt. [lb]')
plt.ylabel('Payload Cost [$/mi]')
plt.title('Payload Cost')


# %%
for i in range(len(free_dwells)):
    if (i % 4 == 0) or i == len(free_dwells) - 1:
        plt.plot(wt_xEV_diffs, time_cost_xy[:,i], 
                 label='{:.2f}'.format(free_dwells[i]))
    else:
        plt.plot(wt_xEV_diffs, time_cost_xy[:,i], linestyle='--')
plt.xlabel('xEV Wt. Minus CV Wt. [lb]')
plt.ylabel('Time Cost [$/mi]')
plt.title('Time Cost')
plt.legend(title='Free Dwell Time [hr]')

# %% [markdown]
# ## Contour plots

# %%
x2d, y2d = np.meshgrid(free_dwells, wt_xEV_diffs / 1000)


# %%
FCS = plt.contourf(x2d, y2d, tot_opp_cost_xy / conv_payload_rev_xy * 100, 20, cmap='jet')
plt.xlabel('Free Dwell Time (hr)')
plt.ylabel('xEV Wt. Minus CV Wt. (1000 lb)')
cb = plt.colorbar(FCS)
cb.set_label('%')
plt.title('Total Opportunity Cost\nPer Conventional Revenue')
plt.show()


# %%
FCS = plt.contourf(x2d, y2d, tot_opp_cost_xy, 20, cmap='jet')
plt.xlabel('Free Dwell Time (hr)')
plt.ylabel('xEV Wt. Minus CV Wt. (1000 lb)')
cb = plt.colorbar(FCS)
cb.set_label('$/Mile')
plt.title('Total Opportunity Cost Per Mile')
plt.show()


# %%
FCS = plt.contourf(x2d, y2d, payload_cost_xy, 15, cmap='jet')
plt.xlabel('Free Dwell Time (hr)')
plt.ylabel('xEV Wt. Minus CV Wt. (1000 lb)')
cb = plt.colorbar(FCS)
cb.set_label("$/Mile")
plt.title('Payload Cost Per Mile')
plt.show()


# %%
FCS = plt.contourf(x2d, y2d, time_cost_xy, 15, cmap='jet')
plt.xlabel('Free Dwell Time (hr)')
plt.ylabel('xEV Wt. Minus CV Wt. (1000 lb)')
cb = plt.colorbar(FCS)
cb.set_label("$/Mile")
plt.title('Time Cost Per Mile')
plt.show()


# %%
FCS = plt.contourf(x2d, y2d, labor_cost_xy, 15, cmap='jet')
plt.xlabel('Free Dwell Time (hr)')
plt.ylabel('xEV Wt. Minus CV Wt. (1000 lb)')
cb = plt.colorbar(FCS)
cb.set_label("$/Mile")
plt.title('Labor Cost Per Mile')
plt.show()


# %%
FCS = plt.contourf(x2d, y2d, payload_time_factor_xy, 15, cmap='jet')
plt.xlabel('Free Dwell Time (hr)')
plt.ylabel('xEV Wt. Minus CV Wt. (1000 lb)')
plt.title('Payload Time Factor')
plt.colorbar(FCS)
plt.show()





# IMPORT Packages and generate inputs for FASTSim (vehicles)

import pandas as pd
import numpy as np
import datetime
import matplotlib.pyplot as plt
from t3co import run_scenario
from t3co import generateinputs

generateinputs.generate()


"""Module containing functions for calculating fuel economy objectives."""

# %%
from __future__ import annotations

from pathlib import Path
from typing import List, Tuple

import fastsim
import matplotlib.pyplot as plt
import numpy as np

from t3co.run import Global as gl
from t3co.run import run_scenario


def get_range_mi(
    mpgge_info: dict, vehicle: fastsim.vehicle.Vehicle, scenario: run_scenario.Scenario
) -> dict:
    """
    This funcion computes range [miles] from mpgge using vehicle powertrain type and energy (or fuel) store size.

    Considerations:
    - at some point each vehicle powertrain type could employ the concept
    of a "first fuel" or "primary fuel" - so return a primary fuel-based range for all
    powertrains.
    - PHEVs have two fuels (generally diesel and electricity). So return two ranges:
    -- One for determining range during optimization
        i.e. the CD range that PHEVs are commonly specified with
        (e.g. PHEV-50 = PHEV with 50 mi AER "All-Electric Range" ~= CD range)
    -- One that represents the "true" total PHEV range (CD + CS using both ESS and FS)

    Args:
        mpgge_info (dict): Dictionary containing MPGGE breakdown
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of analysis vehicle
        scenario (run_scenario.Scenario): Scenario object of current selection

    Returns:
        range_dict (dict): Dictionary containing different range results
    """
    range_dict = {}
    kwh_per_gge = gl.get_kwh_per_gge()
    if vehicle.veh_pt_type == gl.BEV:
        assert (
            vehicle.fs_kwh == 0 and vehicle.fc_max_kw == 0
        ), "Error! BEV vehicle has non-zero ICE attributes - vehicle mass calculation may be off"
        mpgge = mpgge_info["mpgge"]  # use fuel efficiency from battery
        range_mi = (
            vehicle.ess_max_kwh
            * (vehicle.max_soc - vehicle.min_soc)
            * mpgge
            / kwh_per_gge
        )
        range_dict["primary_fuel_range_mi"] = range_mi
    elif vehicle.veh_pt_type == gl.CONV:
        assert (
            vehicle.ess_max_kwh == 0 and vehicle.mc_max_kw == 0
        ), "Error! CONV vehicle has non-zero BEV attributes - vehicle mass calculation may be off"
        mpgge = mpgge_info["mpgge"]
        range_mi = (vehicle.fs_kwh / kwh_per_gge) * mpgge
        range_dict["primary_fuel_range_mi"] = range_mi
    elif vehicle.veh_pt_type == gl.HEV:
        mpgge = mpgge_info["mpgge"]
        elec_range_mi = (
            vehicle.ess_max_kwh
            * (vehicle.max_soc - vehicle.min_soc)
            * mpgge
            / kwh_per_gge
        )
        conv_range_mi = (vehicle.fs_kwh / kwh_per_gge) * mpgge
        range_mi = elec_range_mi + conv_range_mi
        range_dict["primary_fuel_range_mi"] = range_mi
    elif vehicle.veh_pt_type == gl.PHEV:
        # https://github.nrel.gov/AVCI/FASTSim_TCO_Truck/issues/24#issuecomment-39958
        # charge depleting range [miles]
        # CDrangeMiles = MIN( ESSmaxKWh*(CDmaxSOC- CDminSOC)/ CDelectricityKWhperMile , maxFuelStorKWh/33.7/ CDfuelGGEperMile )
        cd_range_mi = min(
            vehicle.ess_max_kwh
            * (vehicle.max_soc - vehicle.min_soc)
            / mpgge_info["cd_electric_kwh__mi"],
            vehicle.fs_kwh / kwh_per_gge * mpgge_info["cd_fuel_mpgge"],
        )
        # charge sustaining range [miles]
        # note, CS range in this way of thinking, is essentially what range is *left over* after you've exhausted the battery to min SOC
        # and switch into CS mode from CD mode, thus we subtract cd_gge_used from the GGE fuel stores of the vehicle
        cd_gge_used = cd_range_mi / mpgge_info["cd_fuel_mpgge"]
        gge_capacity = vehicle.fs_kwh / kwh_per_gge
        cs_range_mi = (gge_capacity - cd_gge_used) * mpgge_info["cs_fuel_mpgge"]
        true_range_mi = cd_range_mi + cs_range_mi
        range_dict = {
            "cd_aer_phev_range_mi": cd_range_mi,
            "cs_phev_range_mi": cs_range_mi,
            "true_phev_range_mi": true_range_mi,
            "primary_fuel_range_mi": cd_range_mi,
        }
    # if sim_drive:
    #     range_dict["cycle_distance_mi"] = sum(sim_drive.cyc.mps * np.diff(sim_drive.cyc.time_s)[0]) * gl.m_to_mi
    #     range_dict["mean_cyc_speed_mph"] = sum(sim_drive.cyc.mps)/max(sim_drive.cyc.time_s) * gl.mps_to_mph
    # elif range_cyc:
    #     range_dict["cycle_distance_mi"] = sum(range_cyc.mps * np.diff(range_cyc.time_s)[0]) * gl.m_to_mi
    #     range_dict["mean_cyc_speed_mph"] = sum(range_cyc.mps)/max(range_cyc.time_s) * gl.mps_to_mph
    return range_dict


def get_sim_drive(erc, v, scenario):
    """
    This helper method returns a FASTSim SimDrive object using the vehicle, drive cycle and scenario

    Args:
        erc (fastsim.cycle.Cycle| List[Tuple[fastsim.cycle.Cycle, float): FASTSim range cycle object or list of tuples of cycles
        v (fastsim.vehicle.Vehicle): FASTSim vehicle object for analysis vehicle
        scenario (run_scenario.Scenario): Scenario object for current selection

    Returns:
        sim_drive (fastsim.simdrive.SimDrive): FASTSim SimDrive object
    """
    sim_drive = fastsim.simdrive.SimDrive(erc, v)
    sim_drive = sim_drive.to_rust()

    sim_params = sim_drive.sim_params
    sim_params.reset_orphaned()
    # sim_params.verbose = False
    if scenario.missed_trace_correction:
        sim_params.missed_trace_correction = True
        sim_params.max_time_dilation = scenario.max_time_dilation
        sim_params.min_time_dilation = scenario.min_time_dilation
        sim_params.time_dilation_tol = scenario.time_dilation_tol
    sim_drive.sim_params = sim_params

    props = sim_drive.props
    props.reset_orphaned()  # see if this is needed
    props.kwh_per_gge = gl.get_kwh_per_gge()
    sim_drive.props = props

    return sim_drive


def get_mpgge(
    eff_range_cyc: fastsim.cycle.Cycle | List[Tuple[fastsim.cycle.Cycle, float]],
    v: fastsim.vehicle.Vehicle,
    scenario,
    diagnostic=False,
) -> Tuple[dict, List[fastsim.simdrive.SimDrive], List[dict]]:
    """
    This helper method gets the composite mpgge fuel efficiency of vehicle for each efficiency_range Drive Cycle and weight.
    It runs the vehicle using efficiency range cycle(s) and returns mpgge based on the powertrain type

    Method computes a composite mpgge from multiple drive cycles and weights for each cycle.
    If the user passes in a single Drive Cycle rather than a list of tuples, the base case of
    a composite mpgge from a single Drive Cycle and a single weight, 1, is computed.

    Also updates the vehicle's corresponding scenario object

    Args:
        eff_range_cyc (fastsim.cycle.Cycle | List[Tuple[fastsim.cycle.Cycle, float]]): efficiency range cycle
        v (fastsim.vehicle.Vehicle): FASTSim vehicle object for analysis vehicle
        scenario (run_scenario.Scenario): Scenario object for current selection
        diagnostic (bool, optional): if True, returns all mpgge dicts. Defaults to False.

    Raises:
        ValueError: unknown vehicle powertrain type

    Returns:
        mpgge_comp (dict): Dictionary containing MPGGE breakdowns
        sim_drives (List[fastsim.simdrive.SimDrive]): List of simdrives for charge depleting and charge sustaining cycles
        mpgges (List[dict], optional): if diagnostic==True, returns additional

    """

    if not isinstance(eff_range_cyc, list):
        eff_range_cyc = [(eff_range_cyc, 1)]

    if v.veh_pt_type == gl.BEV:
        assert (
            v.mc_max_kw > 0
        ), "motor size is 0 kw in BEV - will lead to faulty results"
        assert (
            v.fs_kwh == 0 and v.fc_max_kw == 0
        ), "Error! BEV vehicle has non-zero ICE attributes - vehicle mass calculation may be off"
    elif v.veh_pt_type == gl.CONV:
        assert (
            v.fc_max_kw > 0
        ), "engine size is 0 kw in CONV - will lead to faulty results"
        assert (
            v.ess_max_kwh == 0 and v.mc_max_kw == 0
        ), "Error! CONV vehicle has non-zero BEV attributes - vehicle mass calculation may be off"
    elif v.veh_pt_type == gl.PHEV:
        assert (
            v.fc_max_kw > 0
        ), "engine size is 0 kw in CONV - will lead to faulty results"
        assert (
            v.mc_max_kw > 0
        ), "motor size is 0 kw in BEV - will lead to faulty results"
    elif v.veh_pt_type == gl.HEV:
        pass
    else:
        raise ValueError(f"unknown vehicle powertrain type {v.veh_pt_type}")

    mpgges = []
    weights = []
    sim_drives = []
    for cycle_weight in eff_range_cyc:
        # efficiency-range cycle, weight
        erc, w = cycle_weight

        if v.veh_pt_type == gl.PHEV:
            # get two copies of cycle, rename
            cd_erc = erc.copy()
            cs_erc = erc.copy()
            cd_erc.name = "CD_" + erc.name

            #
            # Charge Depleting Cycle first
            #
            ess_max_kwh_orig = v.ess_max_kwh
            ess_max_kw_orig = v.ess_max_kw
            veh_kg_orig = v.veh_kg
            # infinite battery hack (with no extra weight) to simulate charge depleting behavior
            v.veh_override_kg = veh_kg_orig
            run_scenario.set_max_battery_kwh(v, 50e3)
            run_scenario.set_max_battery_power_kw(v, 1000)
            # don't want extra weight from huge battery
            sim_drive_cd = get_sim_drive(cd_erc, v, scenario)
            assert (
                sim_drive_cd.veh.veh_kg == veh_kg_orig
            ), f"sim_drive.veh.veh_kg == veh_kg_orig / {round(sim_drive_cd.veh.veh_kg)} == {round(veh_kg_orig)}"
            assert (
                sim_drive_cd.veh.ess_max_kwh == v.ess_max_kwh
            ), f"sim_drive.veh.ess_max_kwh == v.ess_max_kwh / {round(sim_drive_cd.veh.ess_max_kwh)} == {round(v.ess_max_kwh)}"
            sim_drive_cd.sim_drive(init_soc=v.max_soc)
            v.veh_override_kg = None

            #
            # Charge Sustaining Cycle second
            #
            # reset battery to inputs for charge sustaining behavior
            run_scenario.set_max_battery_kwh(v, ess_max_kwh_orig)
            run_scenario.set_max_battery_power_kw(v, ess_max_kw_orig)
            # we're supposed to run this as a "hybrid" so set
            # veh_pt_type to gl.HEV so that FASTSim does SOC balancing, run as an HEV, basically
            v.veh_pt_type = gl.HEV
            sim_drive_cs = get_sim_drive(cs_erc, v, scenario)
            sim_drive_cs.sim_drive()

            # CS calcs
            assert (
                sim_drive_cs.props.kwh_per_gge == gl.get_kwh_per_gge()
            ), "fuel LHV not consistent between T3CO and FASTSim"
            cs_fuel_kwh_out_ach = np.sum(sim_drive_cs.fs_kwh_out_ach)
            cs_fuel_kwh__mi = cs_fuel_kwh_out_ach / np.sum(sim_drive_cs.dist_mi)

            # CD calcs
            # sim_drive_cd.fs_kwh_out_ach[i] = sim_drive_cd.fs_kw_out_ach[i] * sim_drive_cd.cyc.dt_s[i] * (1 / 3.6e3)
            assert (
                sim_drive_cd.props.kwh_per_gge == gl.get_kwh_per_gge()
            ), "fuel LHV not consistent between T3CO and FASTSim"
            cd_fuel_kwh_out_ach = np.sum(sim_drive_cd.fs_kwh_out_ach)
            cd_elec_kwh_out_ach = np.sum(
                np.array(sim_drive_cd.ess_kw_out_ach)
                * np.array(sim_drive_cd.cyc.dt_s)
                / 3.6e3
            )
            cd_elec_mpgge = np.sum(sim_drive_cd.dist_mi) / (
                cd_elec_kwh_out_ach / gl.get_kwh_per_gge()
            )
            cd_electric_kwh__mi = cd_elec_kwh_out_ach / np.sum(sim_drive_cd.dist_mi)
            cd_fuel_kwh__mi = cd_fuel_kwh_out_ach / np.sum(sim_drive_cd.dist_mi)

            v.veh_pt_type = gl.PHEV
            v.set_derived()
            assert v.veh_kg == veh_kg_orig
            assert v.ess_max_kwh == ess_max_kwh_orig
            assert v.ess_max_kw == ess_max_kw_orig
            assert v.veh_pt_type == gl.PHEV

            # TODO: some CS sd's have sum(fs_kwh_out_ach) == 0 ... how?
            # These are probably bad designs and should be skipped

            # TODO: these kinds of outputs could be turned into a dataclass to set
            # desired outputs and have stricter checks
            mpgge = {
                #### Charge depleting ####
                "cd_electric_mpgge": cd_elec_mpgge,
                "cd_grid_electric_mpgge": cd_elec_mpgge
                * v.chg_eff,  # can be used for TCO CD mpgge electric
                "cd_fuel_mpgge": sim_drive_cd.mpgge,
                "cd_fuel_mpgde": sim_drive_cd.mpgge / gl.DieselGalPerGasGal,
                "cd_electric_kwh__mi": cd_electric_kwh__mi,
                "cd_fuel_kwh__mi": cd_fuel_kwh__mi,
                "cd_fuel_used_kwh": cd_fuel_kwh_out_ach,
                "cd_elec_used_kwh": cd_elec_kwh_out_ach,
                #### Charge sustaining ####
                "cs_fuel_mpgge": sim_drive_cs.mpgge,
                "cs_fuel_mpgde": sim_drive_cs.mpgge / gl.DieselGalPerGasGal,
                "cs_fuel_kwh__mi": cs_fuel_kwh__mi,
            }
            sim_drives.append(sim_drive_cd)
            sim_drives.append(sim_drive_cs)
        else:
            sim_drive = get_sim_drive(erc, v, scenario)
            sim_drive.sim_drive()

            # FASTSim BEV and PHEV have two different fuel efficiencies, the mpgge from battery and the mpgge from grid
            # grid_mpgge is really just mpgge multiplied by the charger efficiency fraction
            # grid_mpgge is used for fuel cost calculations
            # TCO fuel cost = vmt / mpgge * $/gge
            # for CONV and HEV
            # this is on an energy basis (mpgge has no electric component if the vehicle is CONV)
            # if vehicle is HEV mpgge = mpgge_from_fs_kwh_used + mpgge_from_battery_kwh_used
            if v.veh_pt_type in [gl.CONV, gl.HEV]:
                mpgge_both = sim_drive.mpgge
            else:
                mpgge_both = (
                    sim_drive.mpgge
                    + (1 / sim_drive.electric_kwh_per_mi) * gl.get_kwh_per_gge()
                )
            mpgge = {
                # combine ICE and electric PT components for overall mpgge
                "mpgge": mpgge_both,
                "grid_mpgge": mpgge_both * sim_drive.veh.chg_eff,
                "mpgde": mpgge_both / gl.DieselGalPerGasGal,
                # TODO, have this checked
                "kwh_per_mi": (
                    sum(sim_drive.fs_kwh_out_ach)
                    + ((sim_drive.roadway_chg_kj + sim_drive.ess_dischg_kj) / 3.6e3)
                )
                / sum(sim_drive.dist_mi),
            }
            sim_drives.append(sim_drive)

        mpgges.append(mpgge)
        weights.append(w)

    if v.veh_pt_type == gl.PHEV:
        cs_fuel_mpgge_i = np.array([m["cs_fuel_mpgge"] for m in mpgges])
        cs_fuel_mpgde_i = np.array([m["cs_fuel_mpgde"] for m in mpgges])
        cs_fuel_kwh__mi_i = np.array([m["cs_fuel_kwh__mi"] for m in mpgges])

        cd_fuel_used_kwh_i = [m["cd_fuel_used_kwh"] for m in mpgges]
        cd_battery_used_kwh_i = [m["cd_elec_used_kwh"] for m in mpgges]
        cd_fuel_mpgge_i = np.array([m["cd_fuel_mpgge"] for m in mpgges])
        cd_fuel_mpgde_i = np.array([m["cd_fuel_mpgde"] for m in mpgges])
        cd_electric_mpgge_i = np.array([m["cd_electric_mpgge"] for m in mpgges])
        cd_grid_electric_mpgge_i = np.array(
            [m["cd_grid_electric_mpgge"] for m in mpgges]
        )
        cd_electric_kwh__mi_i = np.array([m["cd_electric_kwh__mi"] for m in mpgges])
        cd_fuel_kwh__mi_i = np.array([m["cd_fuel_kwh__mi"] for m in mpgges])

        mpgge_comp = {
            # Charge sustaining
            "cs_fuel_mpgge": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    cs_fuel_mpgge_i,
                    out=np.zeros_like(cs_fuel_mpgge_i),
                    where=cs_fuel_mpgge_i != 0,
                    casting="unsafe",
                )
            ),
            "cs_fuel_mpgde": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    cs_fuel_mpgde_i,
                    out=np.zeros_like(cs_fuel_mpgde_i),
                    where=cs_fuel_mpgde_i != 0,
                    casting="unsafe",
                )
            ),
            "cs_fuel_kwh__mi": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    cs_fuel_kwh__mi_i,
                    out=np.zeros_like(cs_fuel_kwh__mi_i),
                    where=cs_fuel_kwh__mi_i != 0,
                    casting="unsafe",
                )
            ),
            # Charge depleting
            "cd_fuel_used_kwh_total": sum(cd_fuel_used_kwh_i),
            "cd_battery_used_kwh": sum(cd_battery_used_kwh_i),
            "cd_fuel_mpgge": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    cd_fuel_mpgge_i,
                    out=np.zeros_like(cd_fuel_mpgge_i),
                    where=cd_fuel_mpgge_i != 0,
                    casting="unsafe",
                )
            ),
            "cd_fuel_mpgde": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    cd_fuel_mpgde_i,
                    out=np.zeros_like(cd_fuel_mpgde_i),
                    where=cd_fuel_mpgde_i != 0,
                    casting="unsafe",
                )
            ),
            "cd_electric_mpgge": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    cd_electric_mpgge_i,
                    out=np.zeros_like(cd_electric_mpgge_i),
                    where=cd_electric_mpgge_i != 0,
                    casting="unsafe",
                )
            ),
            "cd_grid_electric_mpgge": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    cd_grid_electric_mpgge_i,
                    out=np.zeros_like(cd_grid_electric_mpgge_i),
                    where=cd_grid_electric_mpgge_i != 0,
                    casting="unsafe",
                )
            ),
            "cd_electric_kwh__mi": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    cd_electric_kwh__mi_i,
                    out=np.zeros_like(cd_electric_kwh__mi_i),
                    where=cd_electric_kwh__mi_i != 0,
                    casting="unsafe",
                )
            ),
            "cd_fuel_kwh__mi": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    cd_fuel_kwh__mi_i,
                    out=np.zeros_like(cd_fuel_kwh__mi_i),
                    where=cd_fuel_kwh__mi_i != 0,
                    casting="unsafe",
                )
            ),
        }
        # also report: AveCombinedkWhperMile = UF*(CDelectricityKWhperMile + CDfuelKWhpermile) + (1-UF)*CSfuelKWhperMile
        uf = run_scenario.get_phev_util_factor(scenario, v, mpgge)
        mpgge_comp["ave_combined_kwh__mile"] = (
            uf * (mpgge_comp["cd_electric_kwh__mi"] + mpgge_comp["cd_fuel_kwh__mi"])
            + (1 - uf) * mpgge_comp["cs_fuel_kwh__mi"]
        )
    else:
        mpgge_i = np.array([m["mpgge"] for m in mpgges])
        grid_mpgge_i = np.array([m["grid_mpgge"] for m in mpgges])
        mpgde_i = np.array([m["mpgde"] for m in mpgges])
        kwh_per_m_i = np.array([m["kwh_per_mi"] for m in mpgges])
        mpgge_comp = {
            "mpgge": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    mpgge_i,
                    out=np.zeros_like(mpgge_i),
                    where=mpgge_i != 0,
                    casting="unsafe",
                )
            ),
            "grid_mpgge": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    grid_mpgge_i,
                    out=np.zeros_like(grid_mpgge_i),
                    where=grid_mpgge_i != 0,
                    casting="unsafe",
                )
            ),
            "mpgde": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    mpgde_i,
                    out=np.zeros_like(mpgde_i),
                    where=mpgde_i != 0,
                    casting="unsafe",
                )
            ),
            # energy consumption kWh/mi is wt-averaged normally (doesn't require inverting / harmonic averaging)
            "kwh_per_mi": np.sum(weights)
            / np.sum(
                np.divide(
                    weights,
                    kwh_per_m_i,
                    out=np.zeros_like(kwh_per_m_i),
                    where=kwh_per_m_i != 0,
                    casting="unsafe",
                )
            ),
        }

    if diagnostic:
        return mpgge_comp, sim_drives, mpgges
    return mpgge_comp, sim_drives


# %%
if __name__ == "__main__":
    vehicle_input_path = Path(
        gl.T3CO_INPUTS_DIR / "tda_example/TDA_FY22_vehicle_model_assumptions.csv"
    ).resolve()
    scenario_inputs_path = Path(
        gl.T3CO_INPUTS_DIR / "tda_example/TDA_FY22_scenario_assumptions.csv"
    ).resolve()
    # load the generated file of vehicles, drive cycles, and tech targets
    PHEV = 365
    v = run_scenario.get_vehicle(PHEV, vehicle_input_path)
    scenario, cycle = run_scenario.get_scenario_and_cycle(PHEV, scenario_inputs_path)
    # cycle = cycle[2][0]
    cyc_file_path = gl.OPTIMIZATION_DRIVE_CYCLES / "regional_haul.csv"
    cycle = run_scenario.load_design_cycle_from_path(cyc_file_path)

    mpgge_comp, sim_drives, mpgges = get_mpgge(cycle, v, scenario, diagnostic=True)
    get_range_mi(mpgge_comp, v, scenario)
    print("phev_utility_factor_override", scenario.phev_utility_factor_override)
    print("phev_utility_factor_computed", scenario.phev_utility_factor_computed)

    fig, (ax1, ax2, ax3) = plt.subplots(3, figsize=(9, 11))

    x = range(0, len(cycle.time_s))
    ax1.tick_params(axis="x", colors="white")
    ax1.tick_params(axis="y", colors="white")
    ax1.plot(x, cycle.mps, label="cycle MPS")
    ax1.plot(x, sim_drives[0].mps_ach, label="MPS ach")
    ax1.set_title("drive cycle MPS and MPS achieved", color="white")
    fig.legend()

    x = range(0, len(cycle.time_s))
    ax2.tick_params(axis="x", colors="white")
    ax2.tick_params(axis="y", colors="white")
    ax2.plot(x, sim_drives[0].mc_mech_kw_out_ach, label="CD mc_mech_kw_out_ach ")
    ax2.plot(x, sim_drives[1].mc_mech_kw_out_ach, label="CS mc_mech_kw_out_ach ")
    ax2.set_title("CD & CS motor kw out", color="white")
    fig.legend()

    x = range(0, len(cycle.time_s))
    ax3.tick_params(axis="x", colors="white")
    ax3.tick_params(axis="y", colors="white")
    ax3.plot(x, sim_drives[0].fc_kw_out_ach, label="CD fc_kw_out_ach")
    ax3.plot(x, sim_drives[1].fc_kw_out_ach, label="CS fc_kw_out_ach")
    ax3.set_title("CD & CS fuel converter kw out", color="white")
    fig.legend()

    # next plot
    fig, (ax1, ax2, ax3, ax4) = plt.subplots(4, figsize=(18, 18))

    CD = 0
    CS = 1

    x = range(0, len(cycle.time_s))
    ax1.tick_params(axis="x", colors="white")
    ax1.tick_params(axis="y", colors="white")
    ax1.plot(
        x,
        sim_drives[0].mc_mech_kw_out_ach,
        dashes=[6, 2],
        label="CD mc_mech_kw_out_ach ",
    )
    ax1.plot(x, sim_drives[0].fc_kw_out_ach, label="CD fc_kw_out_ach")
    ax1.plot(
        x,
        np.array(sim_drives[0].mc_mech_kw_out_ach)
        + np.array(sim_drives[0].fc_kw_out_ach),
        label="CD fc_kw_out_ach + mc_mech_kw_out_ach",
    )
    ax1.set_title("CD motor and FC kw out", color="white")
    fig.legend()

    x = range(0, len(cycle.time_s))
    ax2.tick_params(axis="x", colors="white")
    ax2.tick_params(axis="y", colors="white")
    mk_kwh_out = np.array(sim_drives[0].mc_mech_kw_out_ach) / 3600
    fc_kwh_out = np.array(sim_drives[0].fc_kw_out_ach) / 3600
    mc_kwh_cummultive = np.cumsum(np.array(sim_drives[0].mc_mech_kw_out_ach)) / 3600
    fc_kwh_cummultive = np.cumsum(np.array(sim_drives[0].fc_kw_out_ach)) / 3600
    ax2.plot(
        x, mc_kwh_cummultive, dashes=[6, 2], label="CD MC mech_KWH_out cummulative"
    )
    ax2.plot(x, fc_kwh_cummultive, label="CD FC KWH_out cummulative")
    ax2.plot(
        x,
        mc_kwh_cummultive + fc_kwh_cummultive,
        label=f"CD ALL KWH_out cummulative kwh",
    )
    ax2.set_title(
        f"CD KWH_out cummulative {round(sum(mk_kwh_out + fc_kwh_out))}", color="white"
    )
    fig.legend()

    x = range(0, len(cycle.time_s))
    ax3.tick_params(axis="x", colors="white")
    ax3.tick_params(axis="y", colors="white")
    ax3.plot(
        x,
        sim_drives[1].mc_mech_kw_out_ach,
        dashes=[6, 2],
        label="CS mc_mech_kw_out_ach ",
    )
    ax3.plot(x, sim_drives[1].fc_kw_out_ach, label="CS fc_kw_out_ach")
    ax3.plot(
        x,
        np.array(sim_drives[1].mc_mech_kw_out_ach)
        + np.array(sim_drives[1].fc_kw_out_ach),
        label="CS fc_kw_out_ach + mc_mech_kw_out_ach",
    )
    ax3.set_title("CS motor and FC kw out", color="white")
    fig.legend()

    x = range(0, len(cycle.time_s))
    ax4.tick_params(axis="x", colors="white")
    ax4.tick_params(axis="y", colors="white")
    mk_kwh_out = np.array(sim_drives[1].mc_mech_kw_out_ach) / 3600
    fc_kwh_out = np.array(sim_drives[1].fc_kw_out_ach) / 3600
    mc_kwh_cummultive = np.cumsum(np.array(sim_drives[1].mc_mech_kw_out_ach)) / 3600
    fc_kwh_cummultive = np.cumsum(np.array(sim_drives[1].fc_kw_out_ach)) / 3600
    ax4.plot(
        x, mc_kwh_cummultive, dashes=[6, 2], label="CS MC mech_KWH_out cummulative"
    )
    ax4.plot(x, fc_kwh_cummultive, label="CS FC KWH_out cummulative")
    ax4.plot(
        x,
        mc_kwh_cummultive + fc_kwh_cummultive,
        label=f"CS ALL KWH_out cummulative kwh",
    )
    ax4.set_title(
        f"CS KWH_out cummulative {round(sum(mk_kwh_out + fc_kwh_out))}", color="white"
    )
    fig.legend()

    cd_elec_kwh_out_ach = round(sum(np.array(sim_drives[CD].ess_kw_out_ach) / 3600), 1)
    cd_fuel_kwh_out_ach = round(sum(sim_drives[CD].fs_kwh_out_ach), 1)
    cs_elec_kwh_out_ach = round(sum(np.array(sim_drives[CS].ess_kw_out_ach) / 3600), 1)
    cs_fuel_kwh_out_ach = round(sum(sim_drives[CS].fs_kwh_out_ach), 1)

    print(f"total CD Mode electricity used (from ess_kw_out) {cd_elec_kwh_out_ach}")
    print(f"total CD Mode fuel used (from fs_kwh_out) {cd_fuel_kwh_out_ach}")
    print(f"total CS Mode electricity used (from ess_kw_out) {cs_elec_kwh_out_ach}")
    print(f"total CS Mode fuel used (from fs_kwh_out) {cs_fuel_kwh_out_ach}")
    cd_elec_mpgge = round(
        sum(sim_drives[CD].dist_mi) / (cd_elec_kwh_out_ach / gl.get_kwh_per_gge()), 1
    )
    print(f"CD mode cd_fuel_mpgge {round(sim_drives[CD].mpgge,1)}")
    print(f"CD mode cd_electric_mpgge {cd_elec_mpgge}")
    print(f"CD mode cd_grid_electric_mpgge {cd_elec_mpgge*v.chg_eff}")
    print(f"CS mode cs_fuel_mpgge {round(sim_drives[CS].mpgge,1)}")

    # # %%
    # sel=328
    # selections = [sel]
    # PHEV = 328
    # dir_mark = f"../../run_scripts/PHEV-testing/results/{sel}"
    # vehicles = "../../run_scripts/external_resources/phev-testing/TDA_FY22_vehicle_model_assumptions.csv"
    # scenarios = "../../run_scripts/external_resources/phev-testing/TDA_FY22_scenario_assumptions.csv"
    # Path('../../results/tco-ints/').mkdir(parents=True, exist_ok=True)
    # Path('../../results/tco-results/').mkdir(parents=True, exist_ok=True)
    # gl.TCO_INTERMEDIATES = Path('../../results/tco-ints/')
    # gl.TCO_RESULTS = Path('../../results/tco-results/')
    # gl.write_files = True

    # v = run_scenario.get_vehicle(PHEV, vehicles)
    # scenario, cycle = run_scenario.get_scenario_and_cycle(PHEV, scenarios)
    # cycle = cycle[0][0]

    # mpgge_comp, sim_drives, mpgges = get_mpgge(cycle, v, scenario, diagnostic=True)
    # get_range_mi(mpgge_comp, v, scenario)

    # cd_elec_kwh_out_ach = round(sum(np.array(sim_drives[CD].ess_kw_out_ach) / 3600),1)
    # cd_fuel_kwh_out_ach = round(sum(sim_drives[CD].fs_kwh_out_ach),1)
    # cs_elec_kwh_out_ach = round(sum(np.array(sim_drives[CS].ess_kw_out_ach) / 3600),1)
    # cs_fuel_kwh_out_ach = round(sum(sim_drives[CS].fs_kwh_out_ach),1)
    # print(f'total CD Mode electricity used (from ess_kw_out) {cd_elec_kwh_out_ach}')
    # print(f'total CD Mode fuel used (from fs_kwh_out) {cd_fuel_kwh_out_ach}')
    # print(f'total CS Mode electricity used (from ess_kw_out) {cs_elec_kwh_out_ach}')
    # print(f'total CS Mode fuel used (from fs_kwh_out) {cs_fuel_kwh_out_ach}')
    # print('++++')

    # cd_elec_mpgge  = round(sum(sim_drives[CD].dist_mi) / (cd_elec_kwh_out_ach / gl.get_kwh_per_gge()),1)

    # print(f'CD mode cd_fuel_mpgge {round(sim_drives[CD].mpgge,1)}')
    # print(f'CD mode cd_electric_mpgge {cd_elec_mpgge}')
    # print(f'CD mode cd_grid_electric_mpgge {cd_elec_mpgge*v.chg_eff}')
    # print(f'CS mode cs_fuel_mpgge {round(sim_drives[CS].mpgge,1)}')

# %%



# To add a new cell, type '# %%'
# To add a new markdown cell, type '# %% [markdown]'
# %%
import time
import numpy as np
import matplotlib.pyplot as plt

from dataclasses import asdict
import pandas as pd
import importlib


from t3co import Global as gl
from t3co import moo
importlib.reload(moo)


# %%
RANGE, ACCEL, GRADE, TCO = moo.RANGE, moo.ACCEL, moo.GRADE, moo.TCO


# optimization scenario parameters where we get our baseline vehicle and scenario
# eventually needs parameterization
vnum = 1  #
vocation = "Conv 2020 tech,  750 mi range"
gl.vocation_scenario = vocation
gl.FASTSIM_INPUTS = gl.T2COBENCHMARKDATADIR / 't3cobenchmarkFASTSimInputs.csv'
gl.OTHER_INPUTS = gl.T2COBENCHMARKDATADIR / 't3cobenchmarkOtherInputs.csv'

# can adjust this if we don't get good benchmark/viable results

# sol not found
pop_size = 10
n_max_gen = 15
tol = 0.01

# possible HEV knobs
# light-weighting, fuel converter size, engine efficiency, CdA percent, battery and motor size
knobs_bounds_all = {
    'CdA_perc_imp': (5.0, 25.0),
    'fcMaxOutKw': (200, 500),  # fcMaxOutKw could be used for optimization of HEV
    'ess_max_kwh': (1, 300),
    'mc_max_kw': (10, 300),
    'fc_peak_eff': (46, 60),
    'wt_delta_perc': (0, 25),
}


# %%

for knob, bounds in knobs_bounds_all.items():
    problem = moo.run_optimization(
        pop_size, n_max_gen, tol, {knob: bounds}, vnum, 
        obj_list=[TCO], use_jit=True, optimize_pt=gl.HEV, 
        skip_optimization=True
    )
    xs = np.linspace(bounds[0], bounds[1], 15)

    ys = []

    for x in xs:
        ys.append(problem.get_objs([x]))

    ys = np.array(ys)

    plt.figure()
    plt.plot(xs, ys)
    plt.xlabel(knob)
    plt.ylabel('TCO [$]')


import pandas as pd
import numpy as np
import ast
from pathlib import Path
from t3co.run import Global as gl


def generate(vocation: str, dst: str=gl.OPTIMIZATION_AND_TCO_RCRS):
    """
    This function aggregates specifications from users for powertrains, desired ranges, component costs etc. into two
    csv files - FASTSimInputs and OtherInputs

    Args:
        vocation (str): Vocation type description
        dst (str, optional): results directory file path. Defaults to gl.OPTIMIZATION_AND_TCO_RCRS.
    """

    print("generating inputs")

    vehiclesdir = "vehicles"
    srcdir = "resources"

    BaselineVehicle_path = (
        Path(__file__).parent
        / srcdir
        / vehiclesdir
        / vocation
        / "specifications/BaselineVehicle.csv"
    )
    OptimizerInitializationValues_path = (
        Path(__file__).parent
        / srcdir
        / vehiclesdir
        / vocation
        / "specifications/OptimizerInitializationValues.csv"
    )
    PowertrainTechTargets_path = (
        Path(__file__).parent
        / srcdir
        / vehiclesdir
        / vocation
        / "specifications/PowertrainTechTargets.csv"
    )
    VocationRequirements_path = (
        Path(__file__).parent
        / srcdir
        / vehiclesdir
        / vocation
        / "specifications/VocationRequirements.csv"
    )
    FastsimHeader_path = Path(__file__).parent / "resources" / "auxiliary" / "FASTSimInputsHeader.csv"
    Other_Inputs_Header_path = (
        Path(__file__).parent / "resources" / "auxiliary" / "OtherInputsHeader.csv"
    )

    BaselineVehicleSpec = pd.read_csv(BaselineVehicle_path)
    OptimizerInitializationValues = pd.read_csv(OptimizerInitializationValues_path)
    PowertrainTechTargets = pd.read_csv(PowertrainTechTargets_path)
    VocationRequirements = pd.read_csv(VocationRequirements_path)

    TechYears = BaselineVehicleSpec.loc[:, "Year"].values
    Nyears = len(TechYears)

    Ranges = np.array(ast.literal_eval(VocationRequirements.at[1, "MinRangeMiles"]))
    Nranges = len(Ranges)
    Npowertrains = max(PowertrainTechTargets["PowertrainNumber"].values)

    FASTSimInputsDf = pd.read_csv(FastsimHeader_path)
    OtherInputsDf = pd.read_csv(Other_Inputs_Header_path)

    OtherInputsDf["vmt"] = OtherInputsDf.vmt.astype(str)

    v = 0
    for pt in range(0, Npowertrains):
        for y in range(0, Nyears):
            for r in range(0, Nranges):
                # FASTSimInputsDf.at[v, "scenario_name"] = f"{v/3} {v}"
                FASTSimInputsDf.at[v, "selection"] = v + 1
                FASTSimInputsDf.at[v, "scenario_name"] = (
                    f"{PowertrainTechTargets[PowertrainTechTargets.PowertrainNumber > pt].iloc[0]['PowertrainName']} "
                    f"{BaselineVehicleSpec.at[y, 'Year']} tech, "
                    f" {Ranges[r]} mi range"
                )
                FASTSimInputsDf.at[v, "veh_pt_type"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "veh_pt_type",
                ].values[0]
                FASTSimInputsDf.at[v, "drag_coef"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "drag_coef",
                ]
                FASTSimInputsDf.at[v, "frontalAreaM2"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "frontalAreaM2",
                ]
                FASTSimInputsDf.at[v, "glider_kg"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "glider_kg",
                ]
                FASTSimInputsDf.at[v, "vehCgM"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "vehCgM",
                ]
                FASTSimInputsDf.at[v, "driveAxleWeightFrac"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "driveAxleWeightFrac",
                ]
                FASTSimInputsDf.at[v, "wheelBaseM"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "wheelBaseM",
                ]
                FASTSimInputsDf.at[v, "cargo_kg"] = VocationRequirements.loc[
                    y, "cargo_kg"
                ]
                FASTSimInputsDf.at[v, "vehOverrideKg"] = 0
                FASTSimInputsDf.at[v, "fs_max_kw"] = OptimizerInitializationValues.loc[
                    (OptimizerInitializationValues["PowertrainNumber"] == (pt + 1))
                    & (
                        OptimizerInitializationValues["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fs_max_kw",
                ].values[0]
                FASTSimInputsDf.at[
                    v, "fuelStorSecsToPeakPwr"
                ] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fuelStorSecsToPeakPwr",
                ].values[0]
                FASTSimInputsDf.at[v, "fs_kwh"] = OptimizerInitializationValues.loc[
                    (OptimizerInitializationValues["PowertrainNumber"] == (pt + 1))
                    & (
                        OptimizerInitializationValues["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fs_kwh",
                ].values[0]
                FASTSimInputsDf.at[v, "fs_kwhPerKg"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fs_kwhPerKg",
                ].values[0]
                FASTSimInputsDf.at[v, "fc_max_kw"] = OptimizerInitializationValues.loc[
                    (OptimizerInitializationValues["PowertrainNumber"] == (pt + 1))
                    & (
                        OptimizerInitializationValues["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fc_max_kw",
                ].values[0]
                FASTSimInputsDf.at[v, "fc_eff_type"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fc_eff_type",
                ].values[0]
                FASTSimInputsDf.at[v, "fcAbsEffImpr"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fcAbsEffImpr",
                ].values[0]
                FASTSimInputsDf.at[
                    v, "fuelConvSecsToPeakPwr"
                ] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fuelConvSecsToPeakPwr",
                ].values[0]
                FASTSimInputsDf.at[v, "fuelConvBaseKg"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fuelConvBaseKg",
                ].values[0]
                FASTSimInputsDf.at[v, "fuelConvKwPerKg"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fuelConvKwPerKg",
                ].values[0]
                FASTSimInputsDf.at[v, "mc_max_kw"] = OptimizerInitializationValues.loc[
                    (OptimizerInitializationValues["PowertrainNumber"] == (pt + 1))
                    & (
                        OptimizerInitializationValues["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "mc_max_kw",
                ].values[0]
                # FASTSimInputsDf.at[v, "motorPeakEff"] =                 PowertrainTechTargets.loc[(PowertrainTechTargets['PowertrainNumber'] == (pt+1)) & (PowertrainTechTargets['Year'] == BaselineVehicleSpec.at[y, 'Year']), 'motorPeakEff'].values[0]
                FASTSimInputsDf.at[v, "motorSecsToPeakPwr"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "motorSecsToPeakPwr",
                ].values[0]
                FASTSimInputsDf.at[v, "mcPeKgPerKw"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "mcPeKgPerKw",
                ].values[0]
                FASTSimInputsDf.at[v, "mcPeBaseKg"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "mcPeBaseKg",
                ].values[0]
                FASTSimInputsDf.at[v, "ess_max_kw"] = OptimizerInitializationValues.loc[
                    (OptimizerInitializationValues["PowertrainNumber"] == (pt + 1))
                    & (
                        OptimizerInitializationValues["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "ess_max_kw",
                ].values[0]
                FASTSimInputsDf.at[
                    v, "ess_max_kwh"
                ] = OptimizerInitializationValues.loc[
                    (OptimizerInitializationValues["PowertrainNumber"] == (pt + 1))
                    & (
                        OptimizerInitializationValues["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "ess_max_kwh",
                ].values[0]
                FASTSimInputsDf.at[v, "essKgPerKwh"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "essKgPerKwh",
                ].values[0]
                FASTSimInputsDf.at[v, "essBaseKg"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "essBaseKg",
                ].values[0]
                FASTSimInputsDf.at[v, "essRoundTripEff"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "essRoundTripEff",
                ].values[0]
                FASTSimInputsDf.at[v, "essLifeCoefA"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "essLifeCoefA",
                ].values[0]
                FASTSimInputsDf.at[v, "essLifeCoefB"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "essLifeCoefB",
                ].values[0]
                FASTSimInputsDf.at[v, "wheelInertiaKgM2"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "wheelInertiaKgM2",
                ]
                FASTSimInputsDf.at[v, "numWheels"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "numWheels",
                ]
                FASTSimInputsDf.at[v, "wheelRrCoef"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "wheelRrCoef",
                ]
                FASTSimInputsDf.at[v, "wheelRadiusM"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "wheelRadiusM",
                ]
                FASTSimInputsDf.at[v, "wheelCoefOfFric"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "wheelCoefOfFric",
                ]
                FASTSimInputsDf.at[v, "min_soc"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "min_soc",
                ].values[0]
                FASTSimInputsDf.at[v, "max_soc"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "max_soc",
                ].values[0]
                FASTSimInputsDf.at[
                    v, "essDischgToFcMaxEffPerc"
                ] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "essDischgToFcMaxEffPerc",
                ].values[0]
                FASTSimInputsDf.at[
                    v, "essChgToFcMaxEffPerc"
                ] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "essChgToFcMaxEffPerc",
                ].values[0]
                FASTSimInputsDf.at[v, "maxAccelBufferMph"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "maxAccelBufferMph",
                ].values[0]
                FASTSimInputsDf.at[
                    v, "maxAccelBufferPercOfUseableSoc"
                ] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "maxAccelBufferPercOfUseableSoc",
                ].values[0]
                FASTSimInputsDf.at[v, "percHighAccBuf"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "percHighAccBuf",
                ].values[0]
                FASTSimInputsDf.at[v, "mphFcOn"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "mphFcOn",
                ].values[0]
                FASTSimInputsDf.at[v, "kwDemandFcOn"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "kwDemandFcOn",
                ].values[0]
                FASTSimInputsDf.at[v, "altEff"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "altEff",
                ].values[0]
                FASTSimInputsDf.at[v, "chgEff"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "chgEff",
                ].values[0]
                FASTSimInputsDf.at[v, "auxKw"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "auxKw",
                ]
                FASTSimInputsDf.at[v, "forceAuxOnFC"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "forceAuxOnFC",
                ].values[0]
                FASTSimInputsDf.at[v, "transKg"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "transKg",
                ]
                FASTSimInputsDf.at[v, "transEff"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "transEff",
                ]
                FASTSimInputsDf.at[v, "comp_mass_multiplier"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "comp_mass_multiplier",
                ]
                FASTSimInputsDf.at[v, "essToFuelOkError"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "essToFuelOkError",
                ].values[0]
                FASTSimInputsDf.at[v, "maxRegen"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "maxRegen",
                ].values[0]
                FASTSimInputsDf.at[v, "minFcTimeOn"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "minFcTimeOn",
                ].values[0]
                FASTSimInputsDf.at[v, "idleFcKw"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "idleFcKw",
                ].values[0]
                FASTSimInputsDf.at[v, "fc_eff_map"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "engineMap",
                ].values[0]
                FASTSimInputsDf.at[v, "mcPwrOutPerc"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "mcPwrOutPerc",
                ].values[0]
                FASTSimInputsDf.at[v, "largeBaselineEff"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "largeBaselineEff",
                ].values[0]  # I'm not sure what this is
                FASTSimInputsDf.at[v, "smallBaselineEff"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "smallBaselineEff",
                ].values[0]  # I'm not sure what this is
                FASTSimInputsDf.at[v, "modernMax"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "modernMax",
                ].values[0]  # I'm not sure what this is
                FASTSimInputsDf.at[v, "stopStart"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "stopStart",
                ].values[0]  # I'm not sure what this is

                FASTSimInputsDf.at[v, "valUddsMpgge"] = 0
                FASTSimInputsDf.at[v, "valHwyMpgge"] = 0
                FASTSimInputsDf.at[v, "valCombMpgge"] = 0
                FASTSimInputsDf.at[v, "valUddsKwhPerMile"] = 0
                FASTSimInputsDf.at[v, "valHwyKwhPerMile"] = 0
                FASTSimInputsDf.at[v, "valCombKwhPerMile"] = 0
                FASTSimInputsDf.at[v, "valCdRangeMi"] = 0
                FASTSimInputsDf.at[v, "valConst65MphKwhPerMile"] = 0
                FASTSimInputsDf.at[v, "valConst60MphKwhPerMile"] = 0
                FASTSimInputsDf.at[v, "valConst55MphKwhPerMile"] = 0
                FASTSimInputsDf.at[v, "valConst45MphKwhPerMile"] = 0
                FASTSimInputsDf.at[v, "valUnadjUddsKwhPerMile"] = 0
                FASTSimInputsDf.at[v, "valUnadjHwyKwhPerMile"] = 0
                FASTSimInputsDf.at[v, "val0To60Mph"] = 0
                FASTSimInputsDf.at[v, "valEssLifeMiles"] = 0
                FASTSimInputsDf.at[v, "valRangeMiles"] = 0
                FASTSimInputsDf.at[v, "valVehBaseCost"] = 0
                FASTSimInputsDf.at[v, "valMsrp"] = 0

                # Values to send to OtherInputs file

                OtherInputsDf.at[v, "selection"] = v + 1
                OtherInputsDf.at[v, "vmt"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "vmt",
                ].values[0]
                OtherInputsDf.at[v, "drive_cycle"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "rangeDriveCycleFilePath",
                ].values[0]
                OtherInputsDf.at[v, "segment_name"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "segment_name",
                ].values[0]
                OtherInputsDf.at[v, "gvwr_kg"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "gvwr_kg",
                ].values[0]
                OtherInputsDf.at[v, "gvwr_credit_kg"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "gvwr_credit_kg",
                ].values[0]
                OtherInputsDf.at[v, "fuel_type"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fuel_type",
                ].values[0]
                OtherInputsDf.at[v, "maint_oper_cost_dol_per_mi"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "maint_oper_cost_dol_per_mi",
                ].values[0]
                OtherInputsDf.at[v, "constant_trip_distance_mi"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "constant_trip_distance_mi",
                ].values[0]
                OtherInputsDf.at[v, "vehicle_life_yr"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "vehicle_life_yr",
                ].values[0]
                OtherInputsDf.at[
                    v, "desired_ess_replacements"
                ] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "desired_ess_replacements",
                ].values[0]
                OtherInputsDf.at[v, "discount_rate_pct_per_yr"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "discount_rate_pct_per_yr",
                ]
                OtherInputsDf.at[v, "ess_cost_dol_per_kw"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "ess_cost_dol_per_kw",
                ].values[0]
                OtherInputsDf.at[v, "ess_cost_dol_per_kwh"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "ess_cost_dol_per_kwh",
                ].values[0]
                OtherInputsDf.at[v, "ess_base_cost_dol"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "ess_base_cost_dol",
                ].values[0]
                OtherInputsDf.at[v, "ess_cost_reduction_dol_per_yr"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "ess_cost_reduction_dol_per_yr",
                ].values[0]
                OtherInputsDf.at[v, "ess_salvage_value_dol"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "ess_salvage_value_dol",
                ].values[0]
                OtherInputsDf.at[v, "pe_mc_cost_dol_per_kw"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "pe_mc_cost_dol_per_kw",
                ].values[0]
                OtherInputsDf.at[v, "pe_mc_base_cost_dol"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "pe_mc_base_cost_dol",
                ].values[0]
                OtherInputsDf.at[v, "fc_ice_cost_dol_per_kw"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fc_ice_cost_dol_per_kw",
                ].values[0]
                OtherInputsDf.at[v, "fc_ice_base_cost_dol"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fc_ice_base_cost_dol",
                ].values[0]
                OtherInputsDf.at[v, "fc_fuelcell_cost_dol_per_kw"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fc_fuelcell_cost_dol_per_kw",
                ].values[0]
                OtherInputsDf.at[v, "fs_cost_dol_per_kwh"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fs_cost_dol_per_kwh",
                ].values[0]
                OtherInputsDf.at[v, "fs_h2_cost_dol_per_kwh"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fs_h2_cost_dol_per_kwh",
                ].values[0]
                OtherInputsDf.at[v, "plug_base_cost_dol"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "plug_base_cost_dol",
                ].values[0]
                OtherInputsDf.at[v, "markup_pct"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "markup_pct",
                ]
                OtherInputsDf.at[v, "fc_cng_ice_cost_dol_per_kw"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fc_cng_ice_cost_dol_per_kw",
                ].values[0]
                OtherInputsDf.at[v, "fs_cng_cost_dol_per_kwh"] = PowertrainTechTargets.loc[
                    (PowertrainTechTargets["PowertrainNumber"] == (pt + 1))
                    & (
                        PowertrainTechTargets["Year"]
                        == BaselineVehicleSpec.at[y, "Year"]
                    ),
                    "fs_cng_cost_dol_per_kwh",
                ].values[0]
                OtherInputsDf.at[v, "vehicle_glider_cost_dol"] = BaselineVehicleSpec.loc[
                    PowertrainTechTargets["Year"] == BaselineVehicleSpec.at[y, "Year"],
                    "vehicle_glider_cost_dol",
                ]
                OtherInputsDf.at[v, "tax_rate_pct"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "tax_rate_pct",
                ]
                OtherInputsDf.at[v, "vocation"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "vocation",
                ].values[0]
                OtherInputsDf.at[v, "model_year"] = BaselineVehicleSpec.at[y, "Year"]
                OtherInputsDf.at[v, "region"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "region",
                ].values[0]
                OtherInputsDf.at[v, "target_range_mi"] = Ranges[r]
                OtherInputsDf.at[
                    v, "min_speed_at_6pct_grade_in_5min_mph"
                ] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "min_speed_at_6pct_grade_in_5min_mph",
                ].values[0]
                OtherInputsDf.at[
                    v, "min_speed_at_1p25pct_grade_in_5min_mph"
                ] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "min_speed_at_1p25pct_grade_in_5min_mph",
                ].values[0]
                OtherInputsDf.at[v, "max_time_0_to_60mph_at_gvwr_s"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "max_time_0_to_60mph_at_gvwr_s",
                ].values[0]
                OtherInputsDf.at[v, "max_time_0_to_30mph_at_gvwr_s"] = VocationRequirements.loc[
                    VocationRequirements["Year"] == VocationRequirements.at[y, "Year"],
                    "max_time_0_to_30mph_at_gvwr_s",
                ].values[0]
                # OtherInputsDf.at[v, "Powertrain"] =                  PowertrainTechTargets.loc[(PowertrainTechTargets['PowertrainNumber'] == (pt+1)) & (PowertrainTechTargets['Year'] == BaselineVehicleSpec.at[y, 'Year']), 'PowertrainName'].values[0]

                v += 1

    FASTSimInputsDf.set_index(
        pd.Series(range(1, len(FASTSimInputsDf) + 1)), inplace=True
    )
    FASTSimInputsDf.to_csv(
        dst / gl.FASTSIM_INPUTS_FILE, index=False
    )  # ####Make this into a results folder

    OtherInputsDf.set_index(pd.Series(range(1, len(OtherInputsDf) + 1)), inplace=True)
    OtherInputsDf.to_csv(
        dst / gl.OTHER_INPUTS_FILE, index=False
    )  # ####Make this into a results folder



# todo, clean up * imports
from .run import generateinputs, Global, run_scenario
from .moopack import moo
from .tco import opportunity_cost, tco_analysis, tco_stock_emissions, tcocalc
from .objectives import accel, fueleconomy, gradeability

from . import tests



"""
To run this as a script that drives all the t3co files as a module:
    The working directory for this script should be at the same depth as the t3co/ module.
    Thus the scripts (run_scenario.py, etc.) will have a parent package and you won't get relative import errors.
    ex: run_one.__name__ will be "t3co.run_one"

    First step, generateinputs.generate(vocation)
    Create the files necessary such that run_scenario.run() can read information for Vehicle and Scenario object
    instantiation.

        tldr:
        generateinputs(vocation) -> gl.FASTSIM_INPUTS, gl.OTHER_INPUTS -> run_scenario.run(gl.FASTSIM_INPUTS, gl.OTHER_INPUTS)

        Generate Inputs (GI) first retrieves specific user generated information necessary to create
        two final product files used for TCO and optimization: gl.FASTSIM_INPUTS & gl.OTHER_INPUTS.

        (GI) gets the user generated information via the specified vocation and the following files:
        "t3co/resources/vehicles/{vocation}/specifications/BaselineVehicle.csv"
                                                       "...OptimizerInitializationValues.csv"
                                                       "...PowertrainTechTargets.csv"
                                                       "...VocationRequirements.csv"

        Note: the folder {vocation} must exist and be populated by the user beforehand.

        gl.FASTSIM_INPUTS & gl.OTHER_INPUTS are populated and stored in the default global locations that
        run_scenario.run() uses as default params - implying run() can use other FASTSIM_INPUTS & OTHER_INPUTS files
        if desired.

        Content of gl.FASTSIM_INPUTS and gl.OTHER_INPUTS:

            gl.FASTSIM_INPUTS content:
                # standard FASTSim input file header and rows
                selection,scenario_name,veh_pt_type,drag_coef,frontalAreaM2,glid...
                1.0,"Conv 2020 tech,  750 mi range",1.0,0.546,10.18,11776.0,0...
                2.0,"Conv 2025 tech,  750 mi range",1.0,0.5026666670000001,10...
                3.0,"Conv 2030 tech,  750 mi range",1.0,0.473125,10.18,11776....
                4.0,"Conv 2035 tech,  750 mi range",1.0,0.46,10.18,11776.0,0....

            gl.OTHER_INPUTS content:
                # all information necessary to determine vehicle TCO as well as performance targets
                selection,drive_cycle,vmt_reduct_per_yr,vmt,constant_trip_distance_mi,vehicle_life_yr,...,region,target_range_mi,min_speed_at_6pct_grade_in_5min_mph,min_speed_at_1p25pct_grade_in_5min_mph,max0to60secAtGV
                1.0,long_haul_cyc.csv,0.0,"[100000, 100000, 100000, 100000, 100000, 100000...United States,750.0,30.0,65.0,80.0,20.0
                2.0,long_haul_cyc.csv,,"[100000, 100000, 100000, 100000, 100000, 100000, 1...United States,750.0,30.0,65.0,80.0,20.0
                3.0,long_haul_cyc.csv,,"[100000, 100000, 100000, 100000, 100000, 100000, 1...United States,750.0,30.0,65.0,80.0,20.0
                4.0,long_haul_cyc.csv,,"[100000, 100000, 100000, 100000, 100000, 100000, 1...United States,750.0,30.0,65.0,80.0,20.0
                5.0,long_haul_cyc.csv,,"[100000, 100000, 100000, 100000, 100000, 100000, 1...United States,750.0,30.0,65.0,80.0,20.0

        default values for these files generated by generateinputs():
        gl.FASTSIM_INPUTS: ./t3co/resources/"FASTSimInputs.csv"
        gl.OTHER_INPUTS:   ./t3co/resources/"OtherInputs.csv"

        These files are populated in generateinputs.generate()
        FASTSimInputsDf.to_csv(dst/gl.FASTSIM_INPUTS_FILE, index=False)
        OtherInputsDf.to_csv(dst/gl.OTHER_INPUTS_FILE, index=False)

        where
        gl.FASTSIM_INPUTS_FILE = "FASTSimInputs.csv"
        gl.OTHER_INPUTS = "OtherInputs.csv"
        and
        dst is usually ./t3co/resources

        run() usage:
        run_scenario.run(veh_no, vocation, vehicle_input_path=gl.FASTSIM_INPUTS,
                     scenario_inputs_path=gl.OTHER_INPUTS,

    >>> print('running', __name__)
    >>> generateinputs.generate('Class8_long_haul')
    >>> use_jit = False
    >>> start = time.time()
    >>> res1 = run_scenario.run(1, 'Class8_long_haul', use_jit=use_jit)
    >>> print(f'classic time [s] {time.time()-start} | use jit? {use_jit}')
"""

import time

from t3co import run_scenario
from t3co import generateinputs

if __name__ == "__main__":
    # to regenerate inputs
    # if not Path('t3co/resources/FASTSimInputs.csv').exists():
    #     generateinputs.generate()
    generateinputs.generate('Class8_long_haul')
    use_jit = False
    start = time.time()
    res1 = run_scenario.run(1, 'Class8_long_haul', use_jit=use_jit)
    # res10 = run_one.run_one_scenario(10, use_jit=False)
    # res15 = run_one.run_one_scenario(15, use_jit=False)
    # res22 = run_one.run_one_scenario(22, use_jit=False)
    # res29 = run_one.run_one_scenario(29, use_jit=False)

    print(f'classic time [s] {time.time()-start} | use jit? {use_jit}')

    # start = time.time()
    # jres1 = run_one.run_one_scenario(1, use_jit=True)
    # # jres10 = run_one.run_one_scenario(10, use_jit=True)
    # # jres15 = run_one.run_one_scenario(15, use_jit=True)
    # # jres22 = run_one.run_one_scenario(22, use_jit=True)
    # # jres29 = run_one.run_one_scenario(29, use_jit=True)
    #
    # print(f'jit time [s] {time.time()-start}')
    #
    # print(res1['zero_to_30']==jres1['zero_to_30'])
    # print(res1['zero_to_60_loaded']==jres1['zero_to_60_loaded'])
    # print(res1['mpgge']==jres1['mpgge'])
    # print(res1['grade_6_mph_ach']==jres1['grade_6_mph_ach'])
    # print(res1['veh_cost_set']['msrp']==jres1['veh_cost_set']['msrp'])

    # 'disc_cost': disc_cost,
    # 'discounted_costs_df': discounted_costs_df,
    # 'mpgge': mpgge,
    # 'veh_cost_set': veh_cost_set,
    # 'sim_drive_record': sim_drive_record,
    # 'vehicle': vehicle,
    # 'vehicle_mass': vehicle_mass,
    # 'zero_to_60': zero_to_60,
    # 'zero_to_30': zero_to_30,
    # 'zero_to_60_loaded': zero_to_60_loaded,
    # 'zero_to_30_loaded': zero_to_30_loaded,
    # 'grade_6_mph_ach': grade_6_mph_ach,
    # 'grade_1_25_mph_ach': grade_1_25_mph_ach}


# %%
from __future__ import annotations

import time
from pathlib import Path
from typing import Tuple

import numpy as np
import fastsim
from fastsim import cycle
from fastsim import parameters as params

from t3co.run import Global as gl
from t3co.run import run_scenario


def get_gradeability(
    analysis_vehicle: fastsim.vehicle.Vehicle,
    scenario:run_scenario.Scenario=None,
    verbose:bool=False,
    ess_init_soc:float=None,
    set_weight_to_max_kg:bool=True,
)->Tuple[float, float, fastsim.simdrive.SimDrive, fastsim.simdrive.SimDrive]:
    """
    This function runs SimDrives to determine the gradeability at given speed and the grade vehicle is
    evaluated at how much it meets or exceeds target speed at the target grade.

    Args:
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object for analysis vehicle
        scenario (run_scenario.Scenario, optional): Scenario object for current selection. Defaults to None.
        verbose (bool, optional): if True, prints process steps. Defaults to False.
        ess_init_soc (float, optional): ESS Initial SOC override. Defaults to None.
        set_weight_to_max_kg (bool, optional): if True, run_scenario.set_test_weight() overrides vehice weight to GVWR. Defaults to True.

    Returns:
        grade_6percent_mph_ach (float): Achieved speed on 6% grade test
        grade_1pt25percent_mph_ach (float): Achieved speed on 1.25% grade test
        grade_6_simdrive (fastsim.simdrive.SimDrive): FASTSim SimDrive for gradeability test of 6% grade
        grade_1p25_simdrive (fastsim.simdrive.SimDrive): FASTSim SimDrive for gradeability test of 1.25% grade
    """
    SIX_GRADE = 0.06
    ONE_POINT_TWENTY_FIVE_GRADE = 0.0125
    CYC_SECONDS = 100
    CYC_MPH = 90

    t0 = time.time()

    if scenario is not None and set_weight_to_max_kg:
        kg_before = analysis_vehicle.veh_kg
        run_scenario.set_test_weight(analysis_vehicle, scenario)

    if verbose:
        print(f"f'{Path(__file__).name}:: Vehicle load time: {time.time() - t0:.3f} s")
    # load the cycles

    def get_grade_perf(target_grade: float)-> Tuple[float,fastsim.simdrive.SimDrive]:
        """
        This function obtains the maximum speed achieved on gradeability test for target grade

        Args:
            target_grade (float): Input constant grade for gradeability test

        Returns:
            target_grade_mph_ach (float): Achieved maximum speed mph on gradeability test
            grade_simdrive (fastsim.simdrive.SimDrive): FASTSim SimDrive object for constant grade cycle
        """
        # Test speed achieved at end of 5 minutes at 6% grade (default)
        first_time_step_mph = 0
        if scenario is not None:
            if target_grade == SIX_GRADE:
                first_time_step_mph = scenario.min_speed_at_6pct_grade_in_5min_mph
                if verbose:
                    print(
                        f"f'{Path(__file__).name}:: scenario.min_speed_at_6pct_grade_in_5min_mph: {scenario.min_speed_at_6pct_grade_in_5min_mph}"
                    )
            if target_grade == ONE_POINT_TWENTY_FIVE_GRADE:
                first_time_step_mph = scenario.min_speed_at_1p25pct_grade_in_5min_mph
                if verbose:
                    print(
                        f"f'{Path(__file__).name}:: scenario.min_speed_at_1p25pct_grade_in_5min_mph: {scenario.min_speed_at_1p25pct_grade_in_5min_mph}"
                    )
        if verbose:
            print(
                f"f'{Path(__file__).name}:: first_time_step_mph: {first_time_step_mph}"
            )
        target_grade_cyc_secs = np.arange(CYC_SECONDS)
        cyc_dict = {
            "cycSecs": target_grade_cyc_secs,
            "cycMps": np.append(
                [first_time_step_mph], np.ones(CYC_SECONDS - 1) * CYC_MPH
            )
            / params.MPH_PER_MPS,
            "cycGrade": np.ones(CYC_SECONDS) * target_grade,
        }

        grade_cycle = cycle.Cycle.from_dict(cyc_dict)

        if verbose:
            print(
                f"f'{Path(__file__).name}:: Cycle load time: {time.time() - t0:.3f} s"
            )

        grade_simdrive = run_scenario.get_objective_simdrive(
            analysis_vehicle, grade_cycle
        )

        run_scenario.run_grade_or_accel(
            "grade", analysis_vehicle, grade_simdrive, ess_init_soc
        )

        assert (
            grade_simdrive.trace_miss_dist_frac
            <= grade_simdrive.sim_params.trace_miss_dist_tol
        )

        target_grade_mph_ach = grade_simdrive.mph_ach[-1]
        return target_grade_mph_ach, grade_simdrive

    grade_6percent_mph_ach, grade_6_simdrive = get_grade_perf(SIX_GRADE)
    grade_1pt25percent_mph_ach, grade_1p25_simdrive = get_grade_perf(
        ONE_POINT_TWENTY_FIVE_GRADE
    )
    if scenario is not None and set_weight_to_max_kg:
        run_scenario.reset_vehicle_weight(analysis_vehicle)
        kg_after = analysis_vehicle.veh_kg
        assert (
            kg_after == kg_before
        ), f"total vehicle kg (veh_kg) kg_after must be equal to kg_before: kg_after/kg_before {round(kg_after,4)}/{round(kg_before, 4)}"

    return (
        grade_6percent_mph_ach,
        grade_1pt25percent_mph_ach,
        grade_6_simdrive,
        grade_1p25_simdrive,
    )


# %%

if __name__ == "__main__":
    vehicle_input_path = Path(
        gl.T3CO_INPUTS_DIR / "phev-testing/TDA_FY22_vehicle_model_assumptions.csv"
    ).resolve()
    scenario_inputs_path = Path(
        gl.T3CO_INPUTS_DIR / "phev-testing/TDA_FY22_scenario_assumptions.csv"
    ).resolve()
    # load the generated file of vehicles, drive cycles, and tech targets
    PHEV = 365

    analysis_vehicle = run_scenario.get_vehicle(PHEV, vehicle_input_path)

    scenario, _ = run_scenario.get_scenario_and_cycle(PHEV, scenario_inputs_path)
    (
        grade_6percent_mph_ach,
        grade_1pt25percent_mph_ach,
        grade_6_simdrive,
        grade_1p25_simdrive,
    ) = get_gradeability(
        analysis_vehicle,
        scenario=scenario,
        verbose=True,
        ess_init_soc=None,
        set_weight_to_max_kg=True,
    )
    print(grade_6percent_mph_ach, grade_1pt25percent_mph_ach)
    # %%
    v = analysis_vehicle
    init_socs = [i for i in np.linspace(v.min_soc, v.max_soc, 10)]
    g6mphachs = []
    soc_used = []
    print(f"ess kwh size {v.ess_max_kwh}")
    print(f"max motor kw {v.mc_max_kw}")
    print(f"max fc kw {v.fc_max_kw}")
    for insoc in init_socs:
        g6mphach, g125mphach, g6sd, g125sd = get_gradeability(
            analysis_vehicle,
            scenario=None,
            verbose=False,
            ess_init_soc=insoc,
            set_weight_to_max_kg=False,
        )

        soc_used.append((g6sd.soc[0], g6sd.soc[-1]))

        print(
            "init soc: ",
            round(insoc, 3),
            r"max speed at 6% grade achvd",
            round(g6mphach, 3),
        )

        g6mphachs.append(g6mphach)
# %%



"""Module for simulating acceleration performance."""

# %%
from __future__ import annotations

import time
from pathlib import Path
from typing import Tuple

import numpy as np
from fastsim import cycle
from fastsim import parameters as params

from t3co.run import Global as gl
from t3co.run import run_scenario
import fastsim


def get_accel(
    analysis_vehicle: fastsim.vehicle.Vehicle,
    scenario: run_scenario.Scenario = None,
    set_weight_to_max_kg: bool = True,
    verbose=False,
    ess_init_soc=None,
) -> Tuple[float, float, fastsim.vehicle.Vehicle]:
    """
    This function runs a simdrive for getting 0-to-60 and 0-30 mph time with fully laden weight at GVWR (plus gvwr_credit_kg?)


    Args:
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object for analysis vehicle
        scenario (run_scenario.Scenario, optional): Scenario object for current selection. Defaults to None.
        set_weight_to_max_kg (bool, optional): if True, runs run_scenario.set_test_weight(). Defaults to True.
        verbose (bool, optional): if True, prints the process steps. Defaults to False.
        ess_init_soc (float, optional): ESS initial SOC override. Defaults to None.

    Returns:
        zero_to_sixty (float): 0-60 mph acceleration time in sec
        zero_to_thirty (float): 0-30 mph acceleration time in sec
        accel_simdrive (fastsim.simdrive.SimDrive): FASTSim.simdrive.SimDrive object for running the acceleration drivecycle

    """
    CYC_MPH = 90

    kg_before = None
    if set_weight_to_max_kg and scenario is not None:
        kg_before = analysis_vehicle.veh_kg
        run_scenario.set_test_weight(analysis_vehicle, scenario)

    # load the vehicle
    t0 = time.time()

    if verbose:
        print(f"f'{Path(__file__).name}:: Vehicle load time: {time.time() - t0:.3f} s")
    # load the cycles
    t0 = time.time()
    accel_cyc_secs = np.arange(300)
    cyc_dict = {
        "time_s": accel_cyc_secs,
        "mps": np.append(
            [0], np.ones(len(accel_cyc_secs) - 1) * CYC_MPH / params.MPH_PER_MPS
        ),
        "cycGrade": np.zeros(len(accel_cyc_secs)),
    }

    cyc_accel = cycle.Cycle.from_dict(cyc_dict)

    accel_simdrive = run_scenario.get_objective_simdrive(analysis_vehicle, cyc_accel)

    run_scenario.run_grade_or_accel(
        "accel", analysis_vehicle, accel_simdrive, ess_init_soc
    )
    # print(f'accel:get_accel::>>> {type(accel_simdrive.sim_params)} accel_simdrive.sim_params.trace_miss_dist_tol', accel_simdrive.sim_params.trace_miss_dist_tol)
    # print(f'accel:get_accel::>>> {type(accel_simdrive.sim_params)} accel_simdrive.sim_params.trace_miss_speed_mps_tol ', accel_simdrive.sim_params.trace_miss_speed_mps_tol )
    assert (
        accel_simdrive.trace_miss_dist_frac
        <= accel_simdrive.sim_params.trace_miss_dist_tol
    )

    def test_accel(speed_mph_target: float) -> float:
        """
        This function gets the time it takes to reach target speed

        Args:
            speed_mph_target (float): Target speed for acceleration time

        Returns:
            zero_to_target_mph_s (float): Time taken to reach target speed
        """
        # seconds to target mph test
        if (np.array(accel_simdrive.mph_ach) >= speed_mph_target).any():
            zero_to_target_mph_s = np.interp(
                x=speed_mph_target,
                xp=np.array(accel_simdrive.mph_ach),
                fp=cyc_accel.time_s,
            )
        else:
            # in case vehicle never exceeds speed_mph_target mph, penalize it a lot with a high number
            # print(analysis_vehicle.scenario_name + f' did not achieve {speed_mph_target} mph during the accel test')
            zero_to_target_mph_s = -accel_simdrive.mph_ach[-1]

        return zero_to_target_mph_s

    # seconds to 60 test
    zero_to_sixty = test_accel(60)
    # seconds to 30 test
    zero_to_thirty = test_accel(30)

    #  an assert to ensure that vehicle weight is reset correctly
    if set_weight_to_max_kg and scenario is not None:
        run_scenario.reset_vehicle_weight(analysis_vehicle)
        kg_after = analysis_vehicle.veh_kg
        assert (
            kg_after == kg_before
        ), f"total vehicle kg (veh_kg) kg_after must be to kg_before, {kg_after} != {kg_before}"

    return zero_to_sixty, zero_to_thirty, accel_simdrive


# %%

if __name__ == "__main__":
    # PHEV DEBUGGING
    vehicle_input_path = Path(
        gl.T3CO_INPUTS_DIR / "tda_example/TDA_FY22_vehicle_model_assumptions.csv"
    ).resolve()
    scenario_inputs_path = Path(
        gl.T3CO_INPUTS_DIR / "tda_example/TDA_FY22_scenario_assumptions.csv"
    ).resolve()
    # load the generated file of vehicles, drive cycles, and tech targets
    PHEV = 365
    v = run_scenario.get_vehicle(PHEV, vehicle_input_path)
    scenario, _ = run_scenario.get_scenario_and_cycle(PHEV, scenario_inputs_path)
    zero_to_sixty, zero_to_thirty, accel_simdrive = get_accel(
        v,
        scenario=scenario,
        set_weight_to_max_kg=False,
        verbose=False,
        ess_init_soc=None,
    )
    # %%
    init_socs = [i for i in np.linspace(v.min_soc, v.max_soc, 10)]
    zero_to_sixties = []
    for insoc in init_socs:
        zero_to_sixty, zero_to_thirty, accel_simdrive = get_accel(
            v,
            scenario=scenario,
            set_weight_to_max_kg=False,
            verbose=False,
            ess_init_soc=insoc,
        )
        zero_to_sixties.append(zero_to_sixty)
    # %%



import time
import pandas as pd
from dataclasses import asdict
from fastsim import vehicle

from t3co import Global as gl
from t3co import moo
from t3co.run_scenario import rerun, limit_cargo_kg_for_moo_hev_bev, set_max_battery_kwh, set_max_motor_kw

RANGE, ACCEL, GRADE, TCO = moo.RANGE, moo.ACCEL, moo.GRADE, moo.TCO

if __name__ == '__main__':

    # optimization scenario parameters where we get our baseline vehicle and scenario
    # eventually needs parameterization
    vnum = 1  #
    vocation = "Conv 2020 tech,  750 mi range"
    gl.vocation_scenario = vocation
    gl.FASTSIM_INPUTS = gl.T2COBENCHMARKDATADIR / 't3cobenchmarkFASTSimInputs.csv'
    gl.OTHER_INPUTS = gl.T2COBENCHMARKDATADIR / 't3cobenchmarkOtherInputs.csv'

    # can adjust this if we don't get good benchmark/viable results

    # sol not found
    pop_size = 10
    n_max_gen = 15
    tol = 0.01

    knobs_bounds = {
        'CdA_perc_imp': (5.0, 25.0),
        'ess_max_kwh': (500, 2_000),
        'mc_max_kw': (100, 390),
        'wt_delta_perc': (0, 25),
    }

    t0 = time.time()

    res, problem, solutions, solution = moo.run_optimization(
        pop_size, n_max_gen, knobs_bounds, vnum,
        optimize_pt=gl.BEV, skip_optimization=False, obj_list=[RANGE, ACCEL, GRADE, TCO],
        use_jit=True
    )
    if type(solution) is str and solution == 'solution not found':
        print("ALERT", solution)
    # cProfile.run('res, resd = run_optimization(pop_size, n_max_gen, knobs, knobs_bounds, vnum)',
    #              sort=SortKey.CUMULATIVE
    #              )
    t1 = time.time()

    res_df = solutions
    res_df.to_csv(gl.T2COBENCHMARKDATADIR / "moo_res_df_benchmark_bev.csv")

    def run_veh_for_sweep(i, kwh, motor, conv=False):
        if not conv:
            sweep_vehicle = vehicle.copy_vehicle(problem.mooadvancedvehicle)
            set_max_battery_kwh(sweep_vehicle, kwh)
            set_max_motor_kw(sweep_vehicle, motor)
            limit_cargo_kg_for_moo_hev_bev(problem.opt_scenario, sweep_vehicle)
            problem.opt_scenario.fuel_type = "electricity"
        else:
            problem.opt_scenario.fuel_type = "diesel"
            sweep_vehicle = problem.moobasevehicle
        outdf = rerun(sweep_vehicle, vocation, problem.opt_scenario, use_jit=True)
        outdf["CONV"] = conv
        outdf.update(outdf['veh_msrp_set'])
        outdf.update(outdf['vehicle_mass'])
        if not conv:
            outdf["mpgge"] = outdf["mpgge"]["mpgge"]
        del outdf['veh_msrp_set']
        del outdf['vehicle_mass']
        del outdf['discounted_costs_df']
        del outdf['sim_drive_record']
        del outdf['vehicle']
        del outdf['scenario']
        outdf = pd.DataFrame(outdf, index=[i])
        outdf['vehicle'] = i
        outdf['battery_kwh'] = kwh
        outdf['motor_kw'] = motor
        outdf.set_index('vehicle', inplace=True)
        return outdf

    # for each solution run its results for motor and battery
    # store each vehicle + mpgge results
    sweep_results = []
    for i, kwh, motor in zip(res_df.index, res_df['knob_ess_max_kwh'], res_df['knob_mc_max_kw']):
        outdf = run_veh_for_sweep(i, kwh, motor)
        sweep_results.append(outdf)
    outdf = run_veh_for_sweep(max(res_df.index) + 1, None, None, True)
    sweep_results.append(outdf)

    scenario = problem.opt_scenario
    # pd.DataFrame(asdict(scenario), index=[0]).to_csv(gl.T2COBENCHMARKDATADIR / "t3co_benchmark_scenario.csv")
    sweep_df = pd.concat(sweep_results)
    sweep_df.sort_values(inplace=True, by='tot_cost')
    sweep_df['GVWRLb'] = gl.kg_to_lbs(scenario.gvwr_kg)
    sweep_df['GVWRCreditLb'] = gl.kg_to_lbs(scenario.gvwr_credit_kg)
    sweep_df['GVWRMaxLb'] = gl.kg_to_lbs(scenario.gvwr_credit_kg) + gl.kg_to_lbs(scenario.gvwr_kg)
    sweep_df.to_csv(gl.T2COBENCHMARKDATADIR / "t2co_moo_sweep_solution_set.csv")





# December 2018
# Stock Model in Python
# C. Hunter

from typing import Tuple
import pandas as pd

from t3co.run import Global as gl


def dropCols(df: pd.DataFrame) -> pd.DataFrame:
    """
    This helper method drops columns if any row contains ['*']

    Args:
        df (pd.DataFrame): Input dataframe

    Returns:
        df (pd.DataFrame): Output dataframe with dropped dummy columns
    """
    if df is not None:
        droplist = [i for i in df.columns if df[i].isin(["*"]).any()]
        df = df.drop(droplist, axis=1)
    return df


def stockModel(
    sales: pd.DataFrame,
    marketShares: pd.DataFrame,
    survival: pd.DataFrame,
    annualTravel: pd.DataFrame,
    fuelSplit: pd.DataFrame,
    fuelEfficiency: pd.DataFrame,
    emissions: pd.DataFrame,
    vehicleCosts: pd.DataFrame = None,
    travelCosts: pd.DataFrame = None,
    fuelCosts: pd.DataFrame = None,
    insuranceCosts: pd.DataFrame = None,
    residualCosts: pd.DataFrame = None,
    downtimeCosts: pd.DataFrame = None,
    write_files: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    This function generates the ownershipCosts dataframe from the dataframes for each cost category

    Args:
        sales (pd.DataFrame): Dataframe of yearly number of vehicles sales
        marketShares (pd.DataFrame): Dataframe of yearly Market Share of selection's vocation per vehicle [veh/veh]
        survival (pd.DataFrame): Dataframe of yearly Surviving vehicle per each vehicle [veh/veh]
        annualTravel (pd.DataFrame): Dataframe of vehicle's vmt: Annual Travel [mi/yr]
        fuelSplit (pd.DataFrame): Dataframe of fraction of travel using each fuel [mi/mi]
        fuelEfficiency (pd.DataFrame): Dataframe of vehicle's yearly average fuel efficiency [mi/gge]
        emissions (pd.DataFrame): Dataframe of vehicle's yearly average emissions
        vehicleCosts (pd.DataFrame, optional): Dataframe of vehicle components costs [dol]. Defaults to None.
        travelCosts (pd.DataFrame, optional): Dataframe of maintenance costs [dol/mi]. Defaults to None.
        fuelCosts (pd.DataFrame, optional): Dataframe of fuel operating costs [dol/gge]. Defaults to None.
        insuranceCosts (pd.DataFrame, optional): Dataframe of yearly insurance costs [dol]. Defaults to None.
        residualCosts (pd.DataFrame, optional): Dataframe of yearly residual costs [dol]. Defaults to None.
        downtimeCosts (pd.DataFrame, optional): Dataframe of yearly downtime costs [dol]. Defaults to None.
        write_files (bool, optional): if True, save vehicleCosts, travelCosts, fuelCosts, insuranceCosts,residualCosts, downtimeCosts . Defaults to False.

    Returns:
        stock (pd.DataFrame): Dataframe of stock model of vehicles in the market
        emissions (pd.DataFrame): Dataframe of total emissions
        ownershipCosts (pd.DataFrame): Dataframe of all ownership costs for given selection

    """
    # drop columns with *
    sales = dropCols(sales)
    marketShares = dropCols(marketShares)
    survival = dropCols(survival)
    annualTravel = dropCols(annualTravel)
    fuelSplit = dropCols(fuelSplit)
    fuelEfficiency = dropCols(fuelEfficiency)
    emissions = dropCols(emissions)
    vehicleCosts = dropCols(vehicleCosts)
    travelCosts = dropCols(travelCosts)
    fuelCosts = dropCols(fuelCosts)
    insuranceCosts = dropCols(insuranceCosts)
    residualCosts = dropCols(residualCosts)
    downtimeCosts = dropCols(downtimeCosts)

    # compute stock
    years = marketShares["Model Year"].unique()
    regionalSales = pd.merge(
        sales, marketShares
    )  # on='Model Year'  # seems to default to cross method
    regionalSales["Sales [veh]"] = (
        regionalSales["Sales [veh]"] * regionalSales["Market Share [veh/veh]"]
    )
    regionalSales = regionalSales.drop(["Market Share [veh/veh]"], axis=1)

    stock = []
    for y in years:
        df = regionalSales[regionalSales["Model Year"] <= y].copy()
        df["Year"] = y
        stock.append(df)

    stock = pd.concat(stock)
    stock["Age [yr]"] = stock["Year"] - stock["Model Year"]
    stock = pd.merge(survival, stock)
    stock["Stock [veh]"] = stock["Sales [veh]"] * stock["Surviving Vehicles [veh/veh]"]
    stock = stock.drop(["Surviving Vehicles [veh/veh]"], axis=1)

    # compute vmt
    stock = pd.merge(stock, annualTravel)
    stock["Travel [mi]"] = stock["Annual Travel [mi/yr]"] * stock["Stock [veh]"]
    # stock['Maintenance Cost [dol/mi]'] = stock['Maintenance Cost [dol/mi]'] * stock['Stock [veh]']
    stock = stock.drop(["Annual Travel [mi/yr]", "Age [yr]"], axis=1)

    # compute energy
    energy = pd.merge(pd.merge(stock, fuelEfficiency), fuelSplit)
    energy["Travel [mi]"] = (
        energy["Travel [mi]"] * energy["Fraction of Travel [mi/mi]"]
    )  # travel [mi] by fuel type
    energy["Energy [gge]"] = energy["Travel [mi]"] / energy["Fuel Efficiency [mi/gge]"]

    df = energy.copy()  # store for ownership costs calculations
    df["Age [yr]"] = df["Year"] - df["Model Year"]

    energy = energy.groupby(
        ["Region", "Vehicle", "Vocation", "Model Year", "Year"], as_index=False
    ).agg({"Energy [gge]": "sum"})
    stock = pd.merge(stock, energy).sort_values(
        by=["Year", "Model Year", "Region", "Vocation", "Vehicle"]
    )
    stock = stock[
        [
            "Year",
            "Region",
            "Vocation",
            "Vehicle",
            "Model Year",
            "Sales [veh]",
            "Stock [veh]",
            "Travel [mi]",
            "Energy [gge]",
        ]
    ]

    # # compute emissions if it specified
    # if emissions is not None: # quickly wrote this on 6/19/19 at 10pm at night. Probably a better way to make this more robust and optional inputs for other cost files as well
    #     emissions = pd.merge(df, emissions)
    #     emissions['Emission [g]'] = emissions['Emission Rate [g/gge]'] * emissions['Energy [gge]']
    #     emissions = emissions.drop(['Emission Rate [g/gge]'], axis=1)
    #     emissions = emissions.groupby(['Year','Region','Vocation','Vehicle','Fuel','Pollutant'], as_index=False).agg({'Stock [veh]' : np.sum, 'Travel [mi]' : np.sum, 'Energy [gge]' : np.sum, 'Emission [g]' : np.sum})

    # compute ownership costs
    vehicleCosts = pd.merge(vehicleCosts, df[df["Age [yr]"] < 0.1])
    vehicleCosts["Cost [$]"] = (
        vehicleCosts["Stock [veh]"] * vehicleCosts["Cost [$/veh]"]
    )
    vehicleCosts = vehicleCosts.drop(["Cost [$/veh]"], axis=1)

    travelCosts = pd.merge(travelCosts, df)

    travelCosts["Cost [$]"] = travelCosts["Travel [mi]"] * travelCosts["Cost [$/mi]"]
    travelCosts = travelCosts.drop(["Cost [$/mi]"], axis=1)

    fuelCosts = pd.merge(fuelCosts, df)
    fuelCosts["Cost [$]"] = fuelCosts["Energy [gge]"] * fuelCosts["Cost [$/gge]"]
    fuelCosts = fuelCosts.drop(["Cost [$/gge]"], axis=1)

    insuranceCosts = pd.merge(insuranceCosts, df)
    insuranceCosts["Cost [$]"] = insuranceCosts["Cost [$/Yr]"]
    insuranceCosts = insuranceCosts.drop(["Cost [$/Yr]"], axis=1)

    residualCosts = pd.merge(residualCosts, df)
    residualCosts["Cost [$]"] = residualCosts["Cost [$/Yr]"]
    residualCosts = residualCosts.drop(["Cost [$/Yr]"], axis=1)

    downtimeCosts = pd.merge(downtimeCosts, df)
    downtimeCosts["Cost [$]"] = downtimeCosts["Cost [$/Yr]"]
    downtimeCosts = downtimeCosts.drop(["Cost [$/Yr]"], axis=1)

    if write_files:
        vehicleCosts.to_csv(gl.TCO_INTERMEDIATES / "vehicle_costs.csv", index=False)
        travelCosts.to_csv(gl.TCO_INTERMEDIATES / "travel_costs.csv", index=False)
        fuelCosts.to_csv(gl.TCO_INTERMEDIATES / "fuel_costs.csv", index=False)
        insuranceCosts.to_csv(gl.TCO_INTERMEDIATES / "insurance_costs.csv", index=False)
        residualCosts.to_csv(gl.TCO_INTERMEDIATES / "residual_costs.csv", index=False)
        downtimeCosts.to_csv(gl.TCO_INTERMEDIATES / "downtime_costs.csv", index=False)

    # ownershipCosts = pd.concat([vehicleCosts, travelCosts, fuelCosts], sort=False)
    ownershipCosts = pd.concat(
        [
            vehicleCosts,
            travelCosts,
            fuelCosts,
            insuranceCosts,
            residualCosts,
            downtimeCosts,
        ]
    )
    ownershipCosts = ownershipCosts.groupby(
        ["Region", "Vehicle", "Vocation", "Model Year", "Year", "Category"],
        as_index=False,
    ).agg({"Cost [$]": "sum"})
    ownershipCosts = ownershipCosts.sort_values(
        by=["Year", "Model Year", "Region", "Vocation", "Vehicle"]
    )[["Year", "Region", "Vocation", "Vehicle", "Model Year", "Category", "Cost [$]"]]
    return stock, emissions, ownershipCosts
